#!/usr/bin/env python
#-*- coding: utf-8 -*-
# @file      : scene_and_targets.py
# @author    : Zhi Liu
# @email     : zhiliu.mind@gmail.com
# @homepage  : http://iridescent.ink
# @date      : Sun Nov 27 2022
# @version   : 0.0
# @license   : The Apache License 2.0
# @note      : 
# 
# The Apache 2.0 License
# Copyright (C) 2013- Zhi Liu
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#

import torch as th
import numpy as np
from torchbox.utils.const import *

from torchbox.utils.image import imresize
from torchbox.base.randomfunc import setseed
from torchsar.simulation.geometry import rectangle, disc


def gpts(scene, n=None, nmax=100, seed=None, device='cpu', verbose=False):
    r"""Generates number of point targets.

    Generates number of point targets.

    Parameters
    ----------
    scene : list or tuple
        Scene Area, [xmin, xmax, ymin, ymax]
    n : int or None
        number of targets, if None, randomly chose from 0 to :attr:`nmax`
    nmax : int
        maximum number of targets, default is 100.
    seed : int, optional
        random seed (the default is None, which different seed every time.)
    device : str, optional
        device
    verbose : bool, optional
        show more log info (the default is False, which means does not show)

    Returns
    -------
    targets : tensor
        [tg1, tg2, ..., tgn], tgi = [x,y]
    """

    if verbose:
        print(scene)
    (xmin, xmax, ymin, ymax) = scene

    setseed(seed, 'torch')

    if n is None:
        n = th.randint(0, nmax, (1,), device=device).item()

    x = th.rand((n, 1), device=device) * (xmax - xmin) + xmin
    y = th.rand((n, 1), device=device) * (ymax - ymin) + ymin

    vx = th.zeros((n, 1), device=device)
    vy = th.zeros((n, 1), device=device)
    ax = th.zeros((n, 1), device=device)
    ay = th.zeros((n, 1), device=device)
    rcs = th.rand((n, 1), device=device)
    # print(rcs)
    # rcs = np.ones(n)
    # rcs[0] = 1

    targets = th.cat((x, y, vx, vy, ax, ay, rcs), 1)

    if verbose:
        print(targets)

    return targets


def gdisc(scene, n=None, nmax=100, centers=None, radius=None, amps=None, dx=None, dy=None, seed=None, verbose=False):
    r"""Generates number of Disc targets.

    Generates Disc targets.

    Parameters
    ----------
    scene : list or tuple
        Scene Area, [xmin,xmax,ymin,ymax]
    n : int
        number of disks
    centers : list, optional
        disk centers (the default is None, which generate randomly)
    radius : list, tuple, optional
        If the type of radius is list, specifies the radius of each disk. 
        If the type of radius is tuple, specifies the radius range of all disks.
        The default is :obj:`None`, which means :attr:`radius` is set to ``(1, 30)`` 
    amps : list, optional
        amplitudes (the default is None, which generate randomly)
    dx : float, optional
        resolution in range (default: {1 / (xmax-xmin)})
    dy : float, optional
        resolution in azimuth (default: {1 / (ymax-ymin)})
    seed : int or None, optional
        random seed (the default is None, which different seed every time.)
    verbose : bool, optional
        show more log info (the default is False, which means does not show)

    Returns
    -------
    targets : lists
        [tg1, tg2, ..., tgn], tgi = [x,y]
    """

    if verbose:
        print(scene)
    (xmin, xmax, ymin, ymax) = scene

    if seed is not None:
        th.seed(seed)

    if n is None:
        n = th.randint(0, nmax, (1,)).item()

    if centers is None:
        x0 = th.randint(xmin, xmax, n) * 1.0
        y0 = th.randint(ymin, ymax, n) * 1.0
    else:
        x0 = centers[0]
        y0 = centers[1]

    n = len(x0)

    if radius is None:
        radiusMin, radiusMax = 1, 20
        r = th.randint(radiusMin, radiusMax, n)

    if type(radius) is tuple:
        radiusMin, radiusMax = radius
        r = th.randint(radiusMin, radiusMax, n)

    if amps is None:
        amps = th.rand(n)

    targets = disc(scene, x0[0], y0[0], r[0], a=amps[0], dx=dx, dy=dy, verbose=False)
    for n in range(1, n):
        target = disc(scene, x0[n], y0[n], r[n], a=amps[n], dx=dx, dy=dy, verbose=False)
        targets = np.concatenate((targets, target), axis=0)

    if verbose:
        print(targets)

    return targets


def grectangle(scene, n, amps=None, h=None, w=None, dx=None, dy=None, seed=None, verbose=False):
    """Generates number of rectangle targets.

    Generates number of rectangle targets.

    Parameters
    ----------
    scene : list or tuple
        Scene Area, [xmin, xmax, ymin, ymax]
    n : int
        number of rectangles
    amps : list, optional
        amplitudes (the default is None, which generate randomly)
    h : list, optional
        height of each rectangle (the default is None, which generate randomly)
    w : list, optional
        width of each rectangle (the default is None, which generate randomly)
    dx : float, optional
        resolution in range (default: {1 / (xmax-xmin)})
    dy : float, optional
        resolution in azimuth (default: {1 / (ymax-ymin)})
    seed : int, optional
        random seed (the default is None, which different seed every time.)
    verbose : bool, optional
        show more log info (the default is False, which means does not show)

    Returns
    -------
    targets : tensor
        [tg1, tg2, ..., tgn], tgi = [x,y]
    """

    if verbose:
        print(scene)
    (xmin, xmax, ymin, ymax) = scene

    if seed is not None:
        setseed(seed, 'torch')

    if amps is None:
        amps = th.rand(n)

    x0 = th.randint(xmin, xmax, n) * 1.0
    y0 = th.randint(ymin, ymax, n) * 1.0
    targets = rectangle(scene, x0[0], y0[0], h, w, a=amps[0], dx=None, dy=None, verbose=False)
    for n in range(1, n):
        target = rectangle(scene, x0[n], y0[n], h, w, a=amps[
                           n], dx=None, dy=None, verbose=False)
        targets = np.concatenate((targets, target), axis=0)

    if verbose:
        print(targets)

    return targets


def dsm2tgs(dsm, scene, bg=None, dsize=None, tsize=None, device='cpu'):
    r"""convert digital scene matrix (or digital surface model?) to targets

    Convert digital scene matrix (or digital surface model?) to targets.

    The digital scene matrix (or digital surface model?) has the size of
    :math:`H×W×C_1`, where :math:`H` is the height (Y axis) of the matrix,
    :math:`W` is the width (X axis) of the matrix and :math:`C_1` is the number
    of attributes (without (x, y) coordinate), such as target height(Z axis),
    target velocity, target class and target intensity.


    Parameters
    ----------
    dsm : numpy.tensor or torch.Tensor
        Digital scene matrix.
    bg : float or None, optional
        Background value, 0:black, 1:white (default: None, no thresholding)
    scene : list or tuple, optional
        Scene area [xmin, xmax, ymin, ymax] (default: None, (-W / 2, W / 2, -H / 2, H / 2))
    dsize : list or tuple, optional
        Digital scene matrix size, resize :attr:`dsm` to dsize (default: None, equals to :attr:`dsm`).
    tsize : list or int, optional
        Target size, (default: None, equals to :math:`C_1+2`).
    device : str, optional
        Device string, such as ``'cpu'``(cpu), ``'cuda:0'``, ``'cuda:1'``.

    Returns
    -------
    targets : torch.Tensor
        Targets lists or tensor with shape :math:`N × C_2`,
        where, :math:`N` is the number of targets,
        :math:`C_2` is the dimension of target attribute (with (x, y) coordinate).
    """

    if type(dsm) is not th.Tensor:
        dsm = th.tensor(dsm, device=device)
    else:
        dsm = dsm.to(device)

    if dsm.dim() < 2:
        raise ValueError("---The input digital scene matrix should has 2 or more dimensions!")
    elif dsm.dim() == 2:
        dsm = dsm.unsqueeze(-1)

    H, W, C1 = dsm.shape

    Lx = scene[1] - scene[0]
    Ly = scene[3] - scene[2]

    Hs, Ws = int(Ly), int(Lx)

    if dsize is None:
        dsize = [Hs, Ws]

    if len(dsize) < 2:
        raise ValueError("---The height and width of digital scene matrix must be given!")

    gH, gW, _ = dsm.shape

    if dsize[0] != gH or dsize[1] != gW:
        dsm = imresize(dsm, oshape=dsize[0:2], odtype=dsm.dtype, preserve_range=True)

    gH, gW, _ = dsm.shape

    Coff = 2
    tgidx = th.ones(gH, gW) > 0 if bg is None else dsm[..., -1] > bg
    nTgs = tgidx.sum().item()
    if tsize is None:
        tsize = [nTgs, C1 + Coff]
    if type(tsize) is int:
        tsize = [nTgs, tsize]

    xmin, xmax, ymin, ymax = scene
    # xx = th.linspace(xmin, xmax, gW, device=device).reshape(1, gW).repeat(gH, 1)
    # yy = th.linspace(ymax, ymin, gH, device=device).reshape(gH, 1).repeat(1, gW)
    xx = th.linspace(xmin, xmax, gW, device=device)
    yy = th.linspace(ymax, ymin, gH, device=device)
    xx, yy = th.meshgrid(xx, yy, indexing='xy')
    targets = th.zeros(tsize, device=device, dtype=dsm.dtype)

    targets[:, 0] = xx[tgidx]
    targets[:, 1] = yy[tgidx]
    targets[:, -C1:] = dsm[tgidx][:, -C1:]

    # n = 0
    # for h in range(gH):
    #     for w in range(gW):
    #         if dsm[h, w, -1] > 0:
    #             targets[n, -C1:] = dsm[h, w, -C1:]
    #             targets[n, 0] = xx[h, w]
    #             targets[n, 1] = yy[h, w]
    #             n += 1
    # tend = time.time()

    return targets


def tgs2dsm(tgs, scene, bg=0, dsize=None, device='cpu'):
    r"""targets to digital scene matrix (or digital surface model?)

    Convert targets to digital scene matrix (or digital surface model?).

    The digital scene matrix (or digital surface model?) has the size of
    :math:`H×W×C_1`, where :math:`H` is the height (Y axis) of the matrix,
    :math:`W` is the width (X axis) of the matrix and :math:`C_1` is the number
    of attributes (without (x, y) coordinate), such as target height(Z axis),
    target velocity, target class and target intensity.

    Parameters
    ----------
    targets : list or torch.Tensor
        Targets lists or tensor with shape :math:`N × C_2`,
        where, :math:`N` is the number of targets,
        :math:`C_2` is the dimension of target attribute (with (x, y) coordinate).
    scene : list or tuple, optional
        Scene area [xmin, xmax, ymin, ymax]
    bg : float, optional
        Background value, 0:black, 1:white (default: {0})
    dsize : list or tuple, optional
        Size of digital scene matrix (H, W) (default: None, the scene is discretized
        with 1m)

    Returns
    -------
    dsm : torch.Tensor
        Digital scene matrix.
    """

    if type(tgs) is not th.Tensor:
        targets = th.tensor(tgs, device=device)
    else:
        targets = tgs.clone().detach()

    index = th.zeros(targets.shape[0:2], device=device, dtype=th.int16)
    targets = targets.to(device)

    N, C2 = targets.shape

    Lx = scene[1] - scene[0]
    Ly = scene[3] - scene[2]

    Coff = 0
    if dsize is None:
        dsize = [int(Ly / 1.), int(Lx / 1.), C2 - Coff]
    else:
        if len(dsize) < 2:
            raise ValueError("---The height and width of digital scene matrix must be given!")
        if dsize[0] is None:
            dsize[0] = int(Ly / 1.)
        if dsize[1] is None:
            dsize[1] = int(Lx / 1.)
        if len(dsize) > 2 and dsize[-1] is None:
            dsize[-1] = C2 - Coff

    dsm = th.ones(dsize, device=device) * bg

    H, W = dsize[0:2]

    dx = Lx / (W * 1.0)
    dy = Ly / (H * 1.0)

    index[:, 0] = (targets[:, 0] - scene[0]) / dx
    index[:, 1] = (scene[3] - targets[:, 1]) / dy
    index[index > 63] = 63
    index[index < 0] = 0

    if len(dsize) == 2:
        for idx, target in zip(index, targets):
            dsm[idx[1], idx[0]] = target[-1]  # W:x, H:y
    else:
        for idx, target in zip(index, targets):
            dsm[idx[1], idx[0], -dsize[2]:] = target[-dsize[2]:]  # W:x, H:y

    return dsm


if __name__ == '__main__':

    targets = [
        [0, 0, 0, 0, 0, 0, 1.],
        [-1, -1, 0, 0, 0, 0, .2],
        [2, -1, 0, 0, 0, 0, .3],
        [1, 2, 0, 0, 0, 0, .8],
        [-1, 1, 0, 0, 0, 0, .4],
    ]

    dsm = tgs2dsm(targets, scene=(-4, 4, -4, 4), bg=0, dsize=(4, 4, 5))
    print(dsm.shape)
    # print(dsm[3, 3, :])
    # print(dsm[5, 5, :])

    targets = dsm2tgs(dsm, scene=(-4, 4, -4, 4), bg=0, dsize=(4, 4, 5), tsize=7)
    print(targets)



def saraxis(pdict, axismod=None):
    r"""generate SAR image axes ticks and strings

    Parameters
    ----------
    pdict : dict
        SAR parameter dictionary.
    axismod : str, optional
        axis mode, supported are ``'Image'`` ``'SceneAbsoluteSlantRange'``, ``'SceneRelativeSlantRange'``,
        ``'SceneAbsoluteGroundRange'``, ``'SceneRelativeGroundRange'``, ``'BeamAreaSceneArea'``
         by default None (``'Image'``)

    Returns
    -------
    list and str
        extent, xlabel string, ylabel string, title string.
    """    

def tgshow(targets, scene, shape, extent=None, cmap=None, isflip=[False, False], labelstr=None, outfile=None, isshow=True):
    r"""show targets

    Show targets in an image.

    Arguments
    --------------
    targets : Tensor
        A tensor contains information of targets.
    scene : tuple or list
        Area of scene [xmin, xmax, ymin, ymax].
    shape : tuple or list
        The shape of the scene.

    Keyword Arguments
    --------------
    outfile : str
        The filename for writting figure (default: None, do not save).
    isshow : bool
        Whether to plot figure (default: True).

    Returns
    --------------
        Tensor -- The final image tensor for show.

    Raises
    --------------
        ValueError -- :attr:`targets` should not be None.
    """

def apshow(Srx, Title=None, cmap=None, extent=None, keepAspectRatio=True, outfile=None, isshow=True):
    r"""Show amplitude and phase.

    Show amplitude and phase.

    Parameters
    ----------
    Srx : Tensor
        Complex-valued tensor.
    Title : str, optional
        The figure title (the default is None, which means
        no title).
    cmap : str, optional
        The colormap (the default is None)
    extent : tuple or list, optional
        [description] (the default is None)
    keepAspectRatio : bool, optional
        [description] (the default is True)
    outfile : str, optional
        The filename for outputing (the default is None, which means not save)
    isshow : bool, optional
        Whether to show the figure (the default is True).
    """

def show_response(Srx, extent=None, title="Figure", keepAspectRatio=True, outfile=None, isshow=True):
    r"""show SAR response

    [description]

    Arguments
    ------------------
    Srx : Tensor
        [description]

    Keyword Arguments
    ------------------
    extent : tuple, list or None
        [description] (default: None)
    title : str
        [description] (default: "Figure")
    keepAspectRatio : bool
        [description] (default: True)
    outfile : str
        [description] (default: None)
    isshow : bool
        [description] (default: True)
    """

def showReImAmplitudePhase(Srx, extent=None, title=None, keepAspectRatio=True):
    r"""[summary]

    [description]

    Parameters
    ----------
    Srx : {[type]}
        [description]
    extent : {[type]}, optional
        [description] (the default is None, which [default_description])
    title : str, optional
        [description] (the default is None, which [default_description])
    keepAspectRatio : bool, optional
        [description] (the default is True, which [default_description])
    """

def show_image(img, Title=None, cmap=None, keepAspectRatio=True, outfile=None, isshow=True):
    ...

def sarshow(Xs, pdict, axismod=None, nrows=None, ncols=None, xlabels=None, ylabels=None, titles=None, figsize=None, outfile=None, **kwargs):
    """_summary_

    Parameters
    ----------
    Xs : tensor, array, list or tuple
        list/tuple of image arrays/tensors, if the type is not list or tuple, wrap it.
    pdict : dict
        SAR parameter dict
    axismod : _type_, optional
        _description_, by default None
    nrows : int, optional
        show in :attr:`nrows` rows, by default None (auto computed).
    ncols : int, optional
        show in :attr:`ncols` columns, by default None (auto computed).
    xlabels : str, optional
        labels of x-axis
    ylabels : str, optional
        labels of y-axis
    titles : str, optional
        titles
    figsize : tuple, optional
        figure size, by default None
    outfile : str, optional
        save image to file, by default None (do not save).
    kwargs : 
        see :func:`matplotlib.pyplot.imshow`
    """

def show_sarimage(SI, pdict, axismod=None, title=None, cmap=None, isimgadj=False, aspect=None, outfile=None, newfig=True, figsize=None):
    r"""[summary]

    [description]

    Arguments:
        SI {[type]} -- [description]
        pdict {[type]} -- [description]

    Keyword Arguments:
        axismod {[type]} -- [description] (default: {None})
        title {[type]} -- [description] (default: {None})
        cmap {[type]} -- [description] (default: {None})
        isimgadj bool -- [description] (default: {False})
        aspect {[type]} -- [description] (default: {None})
        outfile {[type]} -- [description] (default: {None})
        newfig bool -- [description] (default: {True})
        figsize {[type]} -- [description] (default: {None})
    """

def show_sarimage3d(SI, pdict, axismod=None, title=None, cmap=None, isimgadj=False, aspect=None, outfile=None, figsize=None):
    r"""[summary]

    [description]

    Arguments:
        SI {[type]} -- [description]
        sarplat {[type]} -- [description]

    Keyword Arguments:
        axismod {[type]} -- [description] (default: {None})
        title {[type]} -- [description] (default: {None})
        cmap {[type]} -- [description] (default: {None})
        isimgadj bool -- [description] (default: {False})
        aspect {[type]} -- [description] (default: {None})
        outfile {[type]} -- [description] (default: {None})
        figsize {[type]} -- [description] (default: {None})
    """

def imshow(Xs, nrows=None, ncols=None, xlabels=None, ylabels=None, titles=None, figsize=None, outfile=None, **kwargs):
    r"""show images

    This function create an figure and show images in :math:`a` rows and :math:`b` columns.

    Parameters
    ----------
    Xs : tensor, array, list or tuple
        list/tuple of image arrays/tensors, if the type is not list or tuple, wrap it.
    nrows : int, optional
        show in :attr:`nrows` rows, by default None (auto computed).
    ncols : int, optional
        show in :attr:`ncols` columns, by default None (auto computed).
    xlabels : str, optional
        labels of x-axis
    ylabels : str, optional
        labels of y-axis
    titles : str, optional
        titles
    figsize : tuple, optional
        figure size, by default None
    outfile : str, optional
        save image to file, by default None (do not save).
    kwargs : 
        see :func:`matplotlib.pyplot.imshow`

    Returns
    -------
    plt
        plot handle
    """





def slantr2groundr(R, H, Ar, Xc):
    """slant range to ground range

    Convert slant range :math:`R` to ground range :math:`X`.

    Parameters
    ----------
    R : 1d-tensor
        slant range array
    H : float
        sarplat height
    Ar : float
        squint angle (unit, rad) in line geometry

    Returns
    -------
    X : 1d-tensor
        ground range

    """

def slantt2groundr(tr, H, Ar):
    """slant time to ground range

    Convert slant range time :math:`t_r` to ground range :math:`X`.

    Parameters
    ----------
    tr : 1d-tensor
        range time array
    H : float
        sarplat height
    Ar : float
        squint angle (unit, rad) in line geometry

    Returns
    -------
    X : 1d-tensor
        ground range

    """

def groundr2slantr(X, H, Ar, Xc):
    """ground range to slant range

    Convert ground range :math:`R` to slant range :math:`X`.

    Parameters
    ----------
    X : 1d-tensor
        ground range array
    H : float
        sarplat height
    Ar : float
        squint angle (unit, rad) in line geometry

    Returns
    -------
    R : 1d-tensor
        slant range

    """

def groundr2slantt(X, H, Ar):
    """ground range to slant time

    Convert ground range :math:`X` to slant time :math:`t_r`.

    Parameters
    ----------
    X : 1d-tensor
        ground range
    H : float
        sarplat height
    Ar : float
        squint angle (unit, rad) in line geometry

    Returns
    -------
    tr : 1d-tensor
        range time array

    """

def min_slant_range(Rnear, Fsr, Noff):
    """minimum slant range from radar to target

    Compute the minimum slant range from radar to target.

    Parameters
    ----------
    Rnear : float
        The nearest range (start sampling) from radar to the target.
    Fsr : float
        Sampling rate in range direction
    Noff : 1d-tensor
        Offset from the start distance (start sampling) cell.

    Returns
    -------
    r : 1d-tensor
        Minimum slant range of each range cell.
    """

def min_slant_range_with_migration(Rnear, Fsr, Noff, Wl, Vr, fdc):




#!/usr/bin/env python
#-*- coding: utf-8 -*-
# @file      : maximum_contrast.py
# @author    : Zhi Liu
# @email     : zhiliu.mind@gmail.com
# @homepage  : http://iridescent.ink
# @date      : Sun Nov 27 2022
# @version   : 2.0
# @license   : The Apache License 2.0
# @note      : 
# 
# The Apache 2.0 License
# Copyright (C) 2013- Zhi Liu
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#

from __future__ import print_function
import torch as th
import torchbox as tb
import torchsar as ts
from tqdm import tqdm


def mcaf_sm(x, phi=None, niter=10, tol=1e-4, eta=0.1, method='N-MCA', selscat=False, axis=-2, ftshift=True, islog=False):
    r"""Contrast based autofocus

    Maximum-Contrast based Autofocus (MCA)

    Parameters
    ----------
    x : Tensor
        complex image with shape :math:`N×N_a×N_r` or :math:`N×N_a×N_r×2`
    phi : float, optional
        initial value of :math:`\phi` (the default is None, which means zeros)
    niter : int, optional
        number of iterations (the default is 10)
    method : str, optional
        method used to update the phase error
        - ``'FP-MCA'`` --> Fix Point
        - ``'CD-MCA'`` --> Coordinate Descent
        - ``'SU-MCA'`` --> Simultaneous Update
        - ``'N-MCA'`` --> Newton, see [2], [1] has problem when used to small image
        - ``'SN-MCA'`` --> Simplified Newton
        - ``'MN-MCA'`` --> Modified Newton
    selscat : bool, optional
        Select brighter scatters.
    axis : int, optional
        Compensation axis.
    ftshift :  bool, optional
        Does shift zero frequency to center?
    islog :  bool, optional
        Does print log info.


    see [1] Zhang S , Liu Y , Li X . Fast Entropy Minimization Based Autofocusing Technique for ISAR Imaging[J]. IEEE Transactions on Signal Processing, 2015, 63(13):3425-3434.
        [2] Zeng T , Wang R , Li F . SAR Image Autofocus Utilizing Minimum-Entropy Criterion[J]. IEEE Geoence & Remote Sensing Letters, 2013, 10(6):1552-1556.
    """

    if not th.is_complex(x):
        x = th.view_as_complex(x)

    if selscat:
        x = ts.center_dominant_scatters(x)
        x = ts.window_data(x, win=None)

    if x.dim() == 2:
        x = x.unsqueeze(0)

    if axis == -2 or axis == 1:
        axisc, axiso = -2, -1
    if axis == -1 or axis == 2:
        axisc, axiso = -1, -2

    N, Na, Nr = x.size()

    raise(ImportError('Not opened yet!'))


if __name__ == '__main__':

    import matplotlib.pyplot as plt

    matfile = '/mnt/e/ws/github/psar/psar/examples/imaging/data/ALPSRP020160970_Vr7180_R3.mat'
    SI = tb.loadmat(matfile)['SI']

    SI0 = SI
    # sa, ea, sr, er = 3500, 3500 + 2048, 5500, 5500 + 2048
    # sa, ea, sr, er = 4600, 4600 + 1024, 5000, 5000 + 1024
    sa, ea, sr, er = 3000, 3000 + 512, 5000, 5000 + 512
    # sa, ea, sr, er = 3200, 3200 + 256, 5000, 5000 + 256

    SI = SI[sa:ea, sr:er, 0] + 1j * SI[sa:ea, sr:er, 1]

    SI0 = SI0[sa:ea, sr:er, 0] + 1j * SI0[sa:ea, sr:er, 1]

    deg = 7
    method, niter = 'N-MCA', 300

    SI = th.from_numpy(SI)
    SI0 = th.from_numpy(SI0)
    # SI = SI.to('cuda:0')
    SI, pa = ts.mcaf_sm(SI, phi=None, niter=niter, tol=1e-4, eta=0.1, method=method, selscat=False, ftshift=True, islog=True)

    print(pa.shape)
    plt.figure()
    plt.plot(pa[-1])
    plt.grid()
    plt.xlabel('Aperture time (samples)')
    plt.ylabel('Phase (rad)')
    plt.title('Estimated phase error (polynomial degree ' + str(deg) + ')')
    plt.show()

    SI0 = th.abs(SI0)
    SI = th.abs(SI)
    print("ENT:", tb.entropy(SI0))
    print("ENT:", tb.entropy(SI))
    print("MSE", th.sum(SI0 - SI))

    print(SI.shape, SI0.shape)
    SI = tb.mapping(SI)
    SI0 = tb.mapping(SI0)
    print(SI.shape, SI0.shape)

    plt.figure()
    plt.subplot(121)
    plt.imshow(SI0, cmap='gray')
    plt.subplot(122)
    plt.imshow(SI[-1], cmap='gray')
    plt.show()



#!/usr/bin/env python
#-*- coding: utf-8 -*-
# @file      : sar_model.py
# @author    : Zhi Liu
# @email     : zhiliu.mind@gmail.com
# @homepage  : http://iridescent.ink
# @date      : Sun Nov 27 2022
# @version   : 0.0
# @license   : The Apache License 2.0
# @note      : 
# 
# The Apache 2.0 License
# Copyright (C) 2013- Zhi Liu
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#

import pickle as pkl
import torch as th
from torchbox.utils.const import *
from torchbox.dsp import normalsignals as sig
from torchsar.sharing.antenna_pattern import antenna_pattern_azimuth


def sarmodel(pdict, mod='2D1', gdshape=None, device='cpu', islog=False):
    r"""model sar imaging process.

    SAR imaging model

    Parameters
    ----------
    pdict: dict
        SAR platform parameters
    mod: str, optional
        mod type(default: '2D1')

            if mod is '2D1':
                the model will be in vector mode:
                    s = A      g
                    MNx1      MNxHW  HWx1(M: Na, N: Nr, MN << HW)
            if mod is '2D2':
                the model will be in mat mode:
                    S = Aa     G     Be
                    MxN       MxH    HxW   WxN(M: Na, N: Nr, M << H, N << W)
            if mod is '1Da':
                    S = A      G
                    MxW       MxH    HxW(M: Na, N: Nr, M << H)
            if mod is '1Dr':
                    S'    =   A      G'
                    NxH       NxW    WxH(M: Na, N: Nr, N << W)

    gdshape: tuple, optional
        discrete scene size of G (default: None, sarplat.acquisition['SceneArea'], dx=dy=1)
    device: str
        device, default is 'cpu'
    islog : bool
        show log info (default: True)

    Returns
    -------
    A : torch tensor
        Imaging mapping matrix.
    """
    if islog:
        print("===In sarmodel...")

    Vg = pdict['Vg']
    Vr = pdict['Vr']
    Ar = pdict['Ar']
    H = pdict['H']
    V = Vr

    Kr, Tp = pdict['Kr'], pdict['Tp']
    Na, Nr = pdict['EchoSize']
    if pdict['GeometryMode'] == 'SceneGeometry':
        Rc = pdict['Rsc']
        R0 = pdict['Rs0']
        SC = th.tensor(pdict['SceneCenter'], device=device)
    if pdict['GeometryMode'] == 'BeamGeometry':
        Rc = pdict['Rbc']
        R0 = pdict['Rb0']
        SC = th.tensor(pdict['BeamCenter'], device=device)
    Wl, Fc, La = pdict['Wl'], pdict['Fc'], pdict['La']

    Na, Nr = pdict['EchoSize']
    SA = pdict['SceneArea']

    gX = SA[1] - SA[0]
    gY = SA[3] - SA[2]
    Xc, Yc = SC[0].item(), SC[1].item()
    ta = pdict['ta']
    tr = pdict['tr'].reshape(-1, 1)
    if type(ta) is not th.Tensor:
        ta = th.from_numpy(ta)
    if type(tr) is not th.Tensor:
        tr = th.from_numpy(tr)

    raise(ImportError('Not opened yet!'))



def __computeRmn(ta, tr, H, V, Xc, Yc, SA, gH, gW):
    r"""compute range at g(i, j)

    compute range at g(i, j)

    Arguments
    ------------
    m int
        current y coordinate of H
    n int
        current x coordinate of W
    ta {time in azimuth}
        azimuth time
    tr numpy array
        range time
    H int
        height of SAR platform
    V float
        velocity of SAR platform
    Yc float
        center coordinate in Y axis
    Xc float
        center coordinate in X axis
    SA list
        scene area: [xmin, xmax, ymin, ymax]
    H int
        height of scene(Y)
    W int
        width of scene(X)
    """

    xmin = SA[0] + Xc
    xmax = SA[1] + Xc
    ymin = SA[2] + Yc
    ymax = SA[3] + Yc

    yy = th.linspace(ymin, ymax, gH).to(ta.device)
    xx = th.linspace(xmin, xmax, gW).to(ta.device)
    ys, xs = th.meshgrid(yy, xx, indexing='ij')

    R = th.sqrt(xs ** 2 + (ys - V * ta) ** 2 + H ** 2)  # [gY, gX]
    R = R.reshape(1, -1)
    ys = ys.reshape(1, -1)
    xs = xs.reshape(1, -1)
    return R, xs, ys


def load_sarmodel(datafile, mod='AinvA'):
    r"""load sarmodel file

    load sarmodel file (``.pkl``)

    Parameters
    ----------
    datafile : str
        Model data file path.
    mod : str, optional
        Specify load which variable, ``A``, ``invA``, ``AinvA``
        (the default is 'AinvA', which :math:`\bm A`, :math:`{\bm A}^{-1}` )

    Returns
    -------
    A : numpy array
        Imaging mapping matrix.
    invA : numpy array
        Inverse of imaging mapping matrix.

    Raises
    ------
    ValueError
        wrong mod
    """
    print("===reading model file: ", datafile, "...")
    if datafile != "":
        # get map
        f = open(datafile, 'rb')
        # for python2
        if sys.version_info < (3, 1):
            if mod == 'A':
                A = pkl.load(f)
                f.close()
                return A
            if mod == 'invA':
                pkl.load(f)
                invA = pkl.load(f)
                f.close()
                return invA
            if mod == 'AinvA':
                A = pkl.load(f)
                invA = pkl.load(f)
                f.close()
                return A, invA
            if mod == 'AAH':
                A = pkl.load(f)
                pkl.load(f)
                AH = pkl.load(f)
                f.close()
                return A, AH
            if mod == 'AinvAAH':
                A = pkl.load(f)
                invA = pkl.load(f)
                AH = pkl.load(f)
                f.close()
                return A, invA, AH
            f.close()
            raise ValueError("mod: 'A', 'invA', 'AinvA', 'AAH', 'AinvAAH'")

        # for python3
        else:
            if mod == 'A':
                A = pkl.load(f, encoding='latin1')
                f.close()
                return A
            if mod == 'invA':
                pkl.load(f, encoding='latin1')
                invA = pkl.load(f, encoding='latin1')
                f.close()
                return invA
            if mod == 'AinvA':
                A = pkl.load(f, encoding='latin1')
                invA = pkl.load(f, encoding='latin1')
                f.close()
                return A, invA
            if mod == 'AAH':
                A = pkl.load(f, encoding='latin1')
                pkl.load(f, encoding='latin1')
                AH = pkl.load(f, encoding='latin1')
                f.close()
                return A, AH
            if mod == 'AinvAAH':
                A = pkl.load(f, encoding='latin1')
                invA = pkl.load(f, encoding='latin1')
                AH = pkl.load(f, encoding='latin1')
                f.close()
                return A, invA, AH
            f.close()
            raise ValueError("mod: 'A', 'invA', 'AinvA', 'AAH', 'AinvAAH'")
    else:
        return None


def save_sarmodel(A, invA=None, AH=None, datafile='./model.pkl'):
    r"""save model mapping matrix

    save model mapping matrix to a file.


    Parameters
    ----------
    A : numpy array
        Imaging mapping matrix
    invA : numpy array, optional
        Moore - Penorse inverse of A(default: {None})
    AH : numpy array, optional
        The Hermite :math:`{\bm A}^H` of the Imaging mapping matrix :math:`\bm A`
        (the default is None, which does not store)
    datafile : str, optional
        save file path(default: {'./model.pkl'})
    """

    f = open(datafile, 'wb')

    pkl.dump(A, f, 0)
    if invA is not None:
        pkl.dump(invA, f, 0)
    if AH is not None:
        pkl.dump(AH, f, 0)
    f.close()




#!/usr/bin/env python
#-*- coding: utf-8 -*-
# @file      : phase_error_model.py
# @author    : Zhi Liu
# @email     : zhiliu.mind@gmail.com
# @homepage  : http://iridescent.ink
# @date      : Sun Nov 27 2022
# @version   : 2.0
# @license   : The Apache License 2.0
# @note      : 
# 
# The Apache 2.0 License
# Copyright (C) 2013- Zhi Liu
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#

from __future__ import print_function
import math
import torch as th
import torchsar as ts
import torchbox as tb


def convert_ppec(cs, degs=None, mode='fftfreq->2fftfreq'):
    r"""convert polynominal phase error

    Convert polynominal phase error.

    Args:
        cs (Tensor): The polynominal coefficients.
        degs (None, optional): The degrees of polynominal.
        mode (str, optional): ``'fftfreq->2fftfreq'`` or ``'2fftfreq->fftfreq'`` (the default is 'fftfeq->2fftfreq')

    Returns:
        Tensor: Converted phase error tensor.

    Raises:
        ValueError: Not support mode.
    """

    if type(cs) is not th.Tensor:
        cs = th.tensor(cs)
    N, M = cs.shape
    if degs is None:
        degs = list(range(2, M + 2))

    ccs = th.zeros_like(cs)
    if mode in ['fftfreq->2fftfreq', 'FFTFREQ->2FFTFREQ']:
        for m, deg in zip(range(M), degs):
            ccs[:, m] = cs[:, m] / (2**deg)
    elif mode in ['2fftfreq->fftfreq', '2FFTFREQ->FFTFREQ']:
        for m, deg in zip(range(M), degs):
            ccs[:, m] = cs[:, m] * (2**deg)
    else:
        raise ValueError('Not support!')
    return ccs


def ppeaxis(n, norm=True, shift=True, mode='fftfreq', dtype=th.float32, device='cpu'):
    r"""Return the sample axis

    Return the sample frequencies axis.

    Parameters
    ----------
    n : int
        Number of samples.
    norm : bool
        Normalize the frequencies.
    shift : bool
        Does shift the zero frequency to center.
    mode : str or None
        Frequency sampling mode. ``'fftfreq'`` or ``'freq'`` normalize the frequencies in (-1/2, 1/2).
        ``'2fftfreq'`` or ``'2freq'`` normalize the frequencies in (-1, 1). If ``None``, return ``None``.
    dtype : Tensor
        Data type, default is ``th.float32``.
    device : str
        device string, default is ``'cpu'``.

    Returns
    -------
    torch array
        Frequency array with size :math:`n×1`.
    """
    fs = n
    if mode in ['fftfreq', 'FFTFREQ']:
        return tb.fftfreq(n, fs, norm=norm, shift=shift, dtype=dtype, device=device)
    if mode in ['freq', 'FREQ']:
        return tb.freq(n, fs, norm=norm, shift=shift, dtype=dtype, device=device)
    if mode in ['2fftfreq', '2FFTFREQ']:
        return 2 * tb.fftfreq(n, fs, norm=norm, shift=shift, dtype=dtype, device=device)
    if mode in ['2freq', '2FREQ']:
        return 2 * tb.freq(n, fs, norm=norm, shift=shift, dtype=dtype, device=device)
    if mode is None:
        return None


def polype(c, x, deg=None):
    r"""compute phase error with polynominal model

    compute phase error with polynominal model

    Args:
        c (Tensor, optional): The polynominal coefficients matrix :math:`{\bf P} = [{\bf p}_1, {\bf p}_2, \cdots, {\bf p}_N ]^T ∈{\mathbb R}^{N × (M -1)}`,
            where, :math:`{\bf p}_n = [p_2, \cdots, p_M]^T`).
        x (Tensor or None, optional): Axis, such as a vector in range of (-1, 1). If :attr:`x` is None, return :attr:`c` directly.
        deg (tuple or None, optional): The degree of the coefficients, :math:`[p_{\rm min}, p_{\rm_max}]`, default is ``None``, which means ``[2, size(c, 1) + 1]``

    Returns:
        (Tensor): The phase error tensor.

    Raises:
        TypeError: :attr:`c` should not be None!
    """

    if x is None:
        return c
    if c is None:
        return None

    if type(c) is not th.Tensor:
        c = th.tensor(c)
    if type(x) is not th.Tensor:
        x = th.tensor(x)

    if deg is None:
        M = [2, c.size(1) + 1]
    else:
        M = deg
    x = x.reshape(1, th.numel(x))

    if c is not None:
        xs = th.tensor([], device=x.device)
        for m in range(M[0], M[1] + 1):
            xs = th.cat((xs, x ** m), axis=0)
        xs = xs.to(c.device)
        c = th.matmul(c, xs)
    else:
        raise TypeError('c should not be None!')

    return c


def dctpe(c, x, deg=None):
    r"""compute phase error with DCT model

    compute phase error with DCT model

    .. math::
        \phi_{e}(h)=\sum_{p=0}^{P} a(p) \cdot d c t(p) \cdot \cos \left[\frac{\pi\left(2 h_{a}+1\right) p}{2 N}\right]

    where, :math:`a(p)=\left{\begin{array}{ll}1 / \sqrt{N} p=0 \ \sqrt{2 / N} p \neq 0\end{array},\right.` and
    :math:`\left{\begin{array}{l}p=0,1,2, \cdots \cdots \cdots, P \ h_{a}=[0,1,2, \cdots \cdots, N-1]\end{array}\right.`

    Args:
        c (Tensor, optional): The DCT coefficients matrix :math:`{\bf P} = [{\bf p}_1, {\bf p}_2, \cdots, {\bf p}_N ]^T ∈{\mathbb R}^{N × (M + 1)}`,
            where, :math:`{\bf p}_n = [p_0, \cdots, p_M]^T`).
        x (Tensor, optional): Axis, such as :math:`(0, Na)`.
        deg (tuple or None, optional): The degree of the coefficients, :math:`[p_{\rm min}, p_{\rm_max}]`, default is ``None``, which means ``[0, size(c, 1)]``

    Returns:
        (Tensor): The phase error tensor.

    Raises:
        TypeError: :attr:`c` should not be None!
    """

    if deg is None:
        M = [0, c.size(1)]
    else:
        M = deg

    N = len(x)
    x = x.reshape(1, th.numel(x))
    if c is not None:
        xs = th.tensor([], device=x.device)
        for m in range(M[0], M[1]):
            if m == 0:
                a = math.sqrt(1. / N)
            else:
                a = math.sqrt(2. / N)
            xs = th.cat((xs, a * th.cos(tb.PI * (2 * x + 1) * m / 2 / N)), axis=0)
        xs = xs.to(c.device)
        c = th.matmul(c, xs)
    else:
        raise TypeError('c should not be None!')

    return c


def rmlpe(phi, mode='poly', deg=4, ftshift=True):
    r"""remove linear phase error

    Remove linear phase error based on linear fitting, the linear phase error
    will cause image shift.

    Args:
        phi (Tensor): A :math:`N×N_x` phase error tensor with linear trend, where, :math:`N` is the batchsize, :math:`N_x` is the samples numbers in direction x.
        mode (str): model mode, default is polynominal
        deg (int): Polynomial degrees (default 4) to fit the error, once fitted, the term of deg=[0,1] will be removed. If :attr:`deg` is None or lower than 2, then we do not fit the error with polynominal and not remove the linear trend.
        ftshift (bool): Shift zero frequency to center when do fft/ifft/fftfreq? (the default is False)

    Returns:
        phi (tensor): Phase without linear trend.
    """

    if phi.dim() == 1:
        phi = phi.unsqueeze(0)  # N-Na
    Nx = phi.size(1)
    if deg is not None and deg >= 2:
        x = ts.ppeaxis(Nx, norm=True, shift=ftshift, mode='2fftfreq', device=phi.device)
        w = tb.polyfit(x, phi, deg=deg)
        w[0, :] = 0.
        w[1, :] = 0.
        phi = tb.polyval(w, x)
        return phi.t()
    else:
        return phi


class PolyPhaseErrorGenerator(object):
    def __init__(self, carange=None, crrange=None, seed=None):
        """Polynominal phase error generator.

        Args:
            carange (None or tuple or list, optional): List of coefficients range of phase error in azimuth direction.
            crrange (None or tuple or list, optional): List of coefficients range of phase error in range direction.
            seed (None or int, optional): The random seed.
        """
        self.carange = carange
        self.crrange = crrange
        self.seed = seed
        self.ma = len(carange[0]) + 1 if carange is not None else None  # order 2-ma
        self.mr = len(crrange[0]) + 1 if crrange is not None else None  # order 2-mr
        if seed is not None:
            tb.setseed(seed)

    def mkpec(self, n, seed=None):
        """Make phase error

        Args:
            n (int): The number of phase errors.
            seed (None or int, optional): The random seed.

        Returns:
            TYPE: Description
        """
        if seed is not None:
            tb.setseed(seed)

        if self.ma is not None:
            ca = th.zeros(n, self.ma - 1)
            for i, lowhigh in enumerate(zip(self.carange[0], self.carange[1])):
                low, high = lowhigh
                ca[:, i] = th.rand(n) * (high - low) + low
        if self.mr is not None:
            cr = th.zeros(n, self.mr - 1)
            for i, lowhigh in enumerate(zip(self.crrange[0], self.crrange[1])):
                low, high = lowhigh
                cr[:, i] = th.rand(n) * (high - low) + low

        if self.ma is not None and self.mr is None:
            return ca
        if self.mr is not None and self.ma is None:
            return cr
        if self.ma is not None and self.mr is not None:
            return ca, cr


if __name__ == '__main__':

    import matplotlib.pyplot as plt

    n = 5
    seed = 2020
    # seed = None
    ftshift = True
    carange = [[-300, -250, -200, -150, -100, -50], [300, 250, 200, 150, 100, 50]]
    crrange = [[-30, -25, -20, -15, -10, -5], [30, 25, 20, 15, 10, 5]]

    carange = [[-300, -300, -300, -300, -300, -300, -300, -300], [300, 300, 300, 300, 300, 300, 300, 300]]

    ppeg = PolyPhaseErrorGenerator(carange, crrange, seed)

    ca, cr = ppeg.mkpec(n=n, seed=None)

    Na, Nr = 256, 256
    xa = ts.ppeaxis(Na, norm=True, shift=ftshift, mode='2fftfreq')
    xr = ts.ppeaxis(Nr, norm=True, shift=ftshift, mode='2fftfreq')
    pa, pr = ts.polype(ca, xa), ts.polype(cr, xr)

    print(ca, ca.shape)
    print(cr, cr.shape)

    print(pa.shape, pr.shape)

    plt.figure()
    plt.subplot(121)
    for i in range(n):
        plt.plot(pa[i])
    plt.subplot(122)
    for i in range(n):
        plt.plot(pr[i])

    plt.show()



#!/usr/bin/env python
#-*- coding: utf-8 -*-
# @file      : sensor_satellite.py
# @author    : Zhi Liu
# @email     : zhiliu.mind@gmail.com
# @homepage  : http://iridescent.ink
# @date      : Sun Nov 27 2022
# @version   : 0.0
# @license   : The Apache License 2.0
# @note      : 
# 
# The Apache 2.0 License
# Copyright (C) 2013- Zhi Liu
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#

from torchbox.utils.const import *

SENSOR_SATELLITE = {

    'GF3': {
        'Fc': 5.400012e9,  # Hz
        'H': 755000,  # height in m
        'V': 7570.67,  # velocity in m/s
        'Tp': 45.00e-6,  # Range pulse length in seconds
        'Kr': 13.3333e11,  # FM rate of radar pulse (Hz/s)
        'Lr': 1,  # antenna length (range) in m
        'La': 10.0,  # antenna length (azimuth) in m
        'PRF': 1413.8530,  # Hz
        # ADC sampling frequency, can be None: Ts = 1 / (1.2*self._sensor['B'])
        'Fs': 66.66e+6,

        'StateVector': None,
        'PlatCenter': None,  # [x, y, z], None: default --> [0, 0, H]
        'SceneCenter': None,
        # SceneArea:[xmin,xmax,ymin,ymax], unit: m
        'SceneArea': None,
        'EchoSize': [32768, 9288],
        'As': 0.0 * PI / 180.0,  # squint angle in earth curve geometry
        'Ad': None,  # depression angle
        'Aon': 30.62 * PI / 180.0,  # off-nadir angle |\
        'Aba': None,  # antenna azimuth beamwidth
        'Abr': None,  # antenna range beamwidth
        'Name': 'GF3',
    },

    'ERS2': {
        'Fc': 5.30e+9,  # Hz
        'H': 780000,  # height in m
        'V': 7125.0,  # velocity in m/s
        'Tp': 37.12e-6,  # Range pulse length in seconds
        'Kr': 4.18989015e+11,  # FM rate of radar pulse (Hz/s)
        'Lr': 1,  # antenna length (range) in m
        'La': 10.0,  # antenna length (azimuth) in m
        'PRF': 1679.9023438,  # Hz
        # ADC sampling frequency, can be None: Ts = 1 / (1.2*self._sensor['B'])
        'Fs': 18.9599991e+6,

        'StateVector': None,
        'PlatCenter': None,  # [x, y, z], None: default --> [0, 0, H]
        'SceneCenter': None,
        # SceneArea:[xmin,xmax,ymin,ymax], unit: m
        'SceneArea': None,
        'EchoSize': [32768, 5616],
        'As': 0.0 * PI / 180.0,  # squint angle in earth curve geometry
        'Ad': None,  # depression angle
        'Aon': 18.35 * PI / 180.0,  # off-nadir angle |\
        'Aba': None,  # antenna azimuth beamwidth
        'Abr': None,  # antenna range beamwidth
        'Name': 'ERS2',
    },

    'RADARSAT1': {
        'Fc': 5.30e+9,  # Hz
        'H': 793000,  # height in m
        'V': 7062.0,  # velocity in m/s
        'Tp': 41.75e-6,  # Range pulse length in seconds
        'Kr': -7.2135e+11,  # FM rate of radar pulse (Hz/s)
        'Lr': 1.5,  # antenna length (range) in m
        'La': 15.0,  # antenna length (azimuth) in m
        'PRF': 1256.98,  # Hz
        # ADC sampling frequency, can be None: Ts = 1 / (1.2*self._sensor['B'])
        'Fs': 32.32e+6,

        'StateVector': None,
        'PlatCenter': None,  # [x, y, z], None: default --> [0, 0, H]
        'SceneCenter': None,
        # SceneArea:[xmin,xmax,ymin,ymax], unit: m
        'SceneArea': None,
        'EchoSize': [19438, 9288],
        'As': -1.53 * PI / 180.0,  # squint angle in earth curve geometry
        'Ad': None,  # depression angle
        'Aon': 30.64 * PI / 180.0,  # off-nadir angle |\
        'Aba': None,  # antenna azimuth beamwidth
        'Abr': None,  # antenna range beamwidth
        'Name': 'RADARSAT1',
    },

    'ALOS': {
        'Fc': 1.27e+9,  # Hz
        'H': 800000,  # height in m
        'V': 7153.0,  # velocity in m/s
        'Tp': 27e-6,  # Range pulse length in seconds
        'Kr': -28.e6 / 27.0e-6,  # FM rate of radar pulse (Hz/s)
        'Lr': 2.9,  # antenna length (range) in m
        'La': 8.9,  # antenna length (azimuth) in m
        'PRF': 1912.0,  # Hz
        # ADC sampling frequency, can be None: Ts = 1 / (1.2*self._sensor['B'])
        'Fs': 32.e6,

        'StateVector': None,
        'PlatCenter': None,  # [x, y, z], None: default --> [0, 0, H]
        'SceneCenter': None,
        # SceneArea:[xmin,xmax,ymin,ymax], unit: m
        'SceneArea': None,
        # 'EchoSize': [512, 1024],
        'EchoSize': [35345, 12040],
        'As': 0.0 * PI / 180.0,  # squint angle in earth curve geometry
        'Ad': None,  # depression angle
        'Aon': 34.3 * PI / 180.0,  # off-nadir angle |\
        'Aba': None,  # antenna azimuth beamwidth
        'Abr': None,  # antenna range beamwidth
        'Name': 'Space3',
    },

}



def sarmodel(pdict, mod='2D1', gdshape=None, device='cpu', islog=False):
    r"""model sar imaging process.

    SAR imaging model

    Parameters
    ----------
    pdict: dict
        SAR platform parameters
    mod: str, optional
        mod type(default: '2D1')

            if mod is '2D1':
                the model will be in vector mode:
                    s = A      g
                    MNx1      MNxHW  HWx1(M: Na, N: Nr, MN << HW)
            if mod is '2D2':
                the model will be in mat mode:
                    S = Aa     G     Be
                    MxN       MxH    HxW   WxN(M: Na, N: Nr, M << H, N << W)
            if mod is '1Da':
                    S = A      G
                    MxW       MxH    HxW(M: Na, N: Nr, M << H)
            if mod is '1Dr':
                    S'    =   A      G'
                    NxH       NxW    WxH(M: Na, N: Nr, N << W)

    gdshape: tuple, optional
        discrete scene size of G (default: None, sarplat.acquisition['SceneArea'], dx=dy=1)
    device: str
        device, default is 'cpu'
    islog : bool
        show log info (default: True)

    Returns
    -------
    A : torch tensor
        Imaging mapping matrix.
    """

def __computeRmn(ta, tr, H, V, Xc, Yc, SA, gH, gW):
    r"""compute range at g(i, j)

    compute range at g(i, j)

    Arguments
    ------------
    m int
        current y coordinate of H
    n int
        current x coordinate of W
    ta {time in azimuth}
        azimuth time
    tr numpy array
        range time
    H int
        height of SAR platform
    V float
        velocity of SAR platform
    Yc float
        center coordinate in Y axis
    Xc float
        center coordinate in X axis
    SA list
        scene area: [xmin, xmax, ymin, ymax]
    H int
        height of scene(Y)
    W int
        width of scene(X)
    """

def load_sarmodel(datafile, mod='AinvA'):
    r"""load sarmodel file

    load sarmodel file (``.pkl``)

    Parameters
    ----------
    datafile : str
        Model data file path.
    mod : str, optional
        Specify load which variable, ``A``, ``invA``, ``AinvA``
        (the default is 'AinvA', which :math:`\bm A`, :math:`{\bm A}^{-1}` )

    Returns
    -------
    A : numpy array
        Imaging mapping matrix.
    invA : numpy array
        Inverse of imaging mapping matrix.

    Raises
    ------
    ValueError
        wrong mod
    """

def save_sarmodel(A, invA=None, AH=None, datafile='./model.pkl'):
    r"""save model mapping matrix

    save model mapping matrix to a file.


    Parameters
    ----------
    A : numpy array
        Imaging mapping matrix
    invA : numpy array, optional
        Moore - Penorse inverse of A(default: {None})
    AH : numpy array, optional
        The Hermite :math:`{\bm A}^H` of the Imaging mapping matrix :math:`\bm A`
        (the default is None, which does not store)
    datafile : str, optional
        save file path(default: {'./model.pkl'})
    """





def __fdnorm(X, p):
    ...

def af_ffo_sm(x, p=2, niter=10, delta=None, toli=1e-2, tolo=1e-2, ftshift=True, islog=False):
    """Stage-by-Stage minimum entropy

    [1] Morrison Jr, Robert Lee; Autofocus, Entropy-based (2002): Entropy-based autofocus for synthetic aperture radar.

    Parameters
    ----------
    x : Tensor
       Corrupt complex SAR image. N-Na-Nr(complex) or N-Na-Nr-2(real)
    niter : int, optional
       The number of iteration (the default is 10)
    delta : {float or None}, optional
       The change step (the default is None (i.e. PI))
    toli : int, optional
       Tolerance error for inner loop (the default is 1e-2)
    tolo : int, optional
       Tolerance error for outer loop (the default is 1e-2)
    ftshift : bool, optional
       Shift the zero frequency to center? (the default is True)
    islog : bool, optional
       Print log information? (the default is False)
    """

def af_ffo(g, w=None, p=2, niter=10, eta=0.5, tol=1e-2, ftshift=False):




#!/usr/bin/env python
#-*- coding: utf-8 -*-
# @file      : make_sar_params.py
# @author    : Zhi Liu
# @email     : zhiliu.mind@gmail.com
# @homepage  : http://iridescent.ink
# @date      : Sun Nov 27 2022
# @version   : 2.0
# @license   : The Apache License 2.0
# @note      : 
# 
# The Apache 2.0 License
# Copyright (C) 2013- Zhi Liu
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#

from __future__ import print_function
import math
import torch as th
import torchbox as tb


class SARParameterGenerator(object):

    """SAR Parameter Generator

    Attributes
    ----------
    prdict : dict
        ``{'p1': [low, high], 'p2': [low, high]...}``
    seed : int or None
        The seed for random generator.
    """

    def __init__(self, prdict, seed=None):
        self.prdict = prdict
        self.seed = seed
        if seed is not None:
            tb.setseed(seed)

    def mksarp(self, n=1, seed=None):
        """Makes SAR Parameters

        Parameters
        ----------
        n : int, optional
            The number of experiments.
        seed : None, optional
            The seed for random generator.

        Returns
        -------
        dict
            The SAR Parameter dict.
        """
        if seed is not None:
            tb.setseed(seed)

        pdict = {}
        for k, v in self.prdict.items():
            if v is not None:
                low, high = v
                if n == 1:
                    pdict[k] = th.rand(n).item() * (high - low) + low
                else:
                    pdict[k] = th.rand(n) * (high - low) + low

        return pdict


if __name__ == '__main__':

    seed = 2020
    prdict = {}
    prdict['H'] = (0, 100)
    prdict['V'] = (11, 22)

    sarg = SARParameterGenerator(prdict, seed=seed)

    pdict = sarg.mksarp()

    print(pdict)



#!/usr/bin/env python
#-*- coding: utf-8 -*-
# @file      : uavsar.py
# @author    : Zhi Liu
# @email     : zhiliu.mind@gmail.com
# @homepage  : http://iridescent.ink
# @date      : Sun Nov 27 2022
# @version   : 0.0
# @license   : The Apache License 2.0
# @note      : 
# 
# The Apache 2.0 License
# Copyright (C) 2013- Zhi Liu
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#

import os
import re
import struct
import logging
import numpy as np
from torchsar.products.ceos import decfmtfceos, read_ceos_sar_raw, read_ceos_sar_slc
from torchsar.products.record import readrcd1item


# B: number in binary format
# A: string
# I: char
SarDataFileFileDescriptorRecordERS = {
    # SAR DATA FILE - FILE DESCRIPTOR RECORD (FIXED SEGMENT)
    'Record sequence number': [(1, 4), '1B4', 0],
    'l-st record sub-type code': [(5, 5), '1B1', 0],
    'Record type code': [(6, 6), '1B1', 0],
    '2-nd record sub-type code': [(7, 7), '1B1', 0],
    '3-rd record sub-type code': [(8, 8), '1B1', 0],
    'Length of this record': [(9, 12), '1B4', 0],
    'ASCII/EBCDIC flag': [(13, 14), '1A2', 0],
    '1Blanks': [(15, 16), '1A2', 0],
    'Format control document ID': [(17, 28), '1A12', 0],
    'Format control document revision level': [(29, 30), '1A2', 0],
    'File design descriptor revision letter': [(31, 32), '1A2', 0],
    'Generating software release and revision level': [(33, 44), '1A12', 0],
    'File number': [(45, 48), '1I4', 0],
    'File name': [(49, 64), '1A16', 0],
    'Record sequence and location type flag': [(65, 68), '1A4', 0],
    'Sequence number location': [(69, 76), '1I8', 0],
    'Sequence number field length': [(77, 80), '1I4', 0],
    'Record code and location type flag': [(81, 84), '1A4', 0],
    'Record code location': [(85, 92), '1I8', 0],
    'Record code field length': [(93, 96), '1I4', 0],
    'Record length and location type flag': [(97, 100), '1A4', 0],
    'Record length location': [(101, 108), '1I8', 0],
    'Record length field length': [(109, 112), '1I4', 0],
    'Reserved1': [(113, 113), '1I1', 0],
    'Reserved2': [(114, 114), '1I1', 0],
    'Reserved3': [(115, 115), '1I1', 0],
    'Reserved4': [(116, 116), '1I1', 0],
    'Reserved segment': [(117, 180), '1A64', 0],
    # SAR DATA IMAGERY OPTIONS FILE DESCRIPTOR RECORD (VARIABLE SEGMENT)
    'Number of SAR DATA records (nominal)': [(181, 186), '1I6', 0],
    'SAR DATA record length (bytes)': [(187, 192), '1I6', 0],
    'Reserved1 (blanks)': [(193, 216), '1A4', 0],
    'Number of bits per sample': [(217, 220), '1I4', 0],
    'Number of samples per data group (or pixels)': [(221, 224), '1I4', 0],
    'Number of bytes per data group (or pixel)': [(225, 228), '1I4', 0],
    'Justification and order of samples within data group': [(229, 232), '1A4', 0],
    'Number of SAR channels in this file': [(233, 236), '1I4', 0],
    'Number of lines per data set (nominal)': [(237, 244), '1I8', 0],
    'Number of left border pixels per line': [(245, 248), '1I4', 0],
    'Total number of data groups per line per SAR channel': [(249, 256), '1I8', 0],
    'Number of right border pixels per line': [(257, 260), '1I4', 0],
    'Number of top border lines': [(261, 264), '1I4', 0],
    'Number of bottom border lines': [(265, 268), '1I4', 0],
    'Interleaving indicator': [(269, 272), '1A4', 0],
    'Number of physical records per line': [(273, 274), '1I2', 0],
    'Number of physical records per multi-channel line': [(275, 276), '1I2', 0],
    'Number of bytes of prefix data per record': [(277, 280), '1I4', 0],
    'Number of bytes of SAR data (or pixel data) per record (nominal)': [(281, 288), '1I8', 0],
    'Number of bytes of suffix data per record': [(289, 292), '1I4', 0],
    'Reserved2': [(293, 340), '1A48', 0],
    'Blanks': [(341, 368), '1A28', 0],
    'Reserved3': [(369, 400), '1A32', 0],
    'SAR Data format type identifier': [(401, 428), '1A28', 0],
    'SAR Data format type code': [(429, 432), '1A4', 0],
    'Number of left fill bits within pixel': [(433, 436), '1I4', 0],
    'Number of right fill bits within pixel': [(437, 440), '1I4', 0],
    'Maximum data range of pixel (max-min value for I and Q)': [(441, 448), '1I8', 0],
    # 'spare': [(449, 11644), '11196s', 0],
}

SarDataFileSignalDataRecordERS = {
    'Record sequence number': [(1, 4), '1B4', 0],
    'l-st record sub-type code': [(5, 5), '1B1', 0],
    'Record type code': [(6, 6), '1B1', 0],
    '2-nd record sub-type code': [(7, 7), '1B1', 0],
    '3-rd record sub-type code': [(8, 8), '1B1', 0],
    'Length of this record': [(9, 12), '1B4', 0],
    # PREFIX DATA - GENERAL INFORMATION
    'SAR image data line number': [(13, 16), '1B4', 0],
    'SAR image data record index (indicates the record sequence number of the image line)': [(17, 20), '1B4', 0],
    'Actual count of left-fill pixels': [(21, 24), '1B4', 0],
    'Actual count of data pixels (samples)': [(25, 28), '1B4', 0],
    'Actual count of right-fill pixels': [(29, 32), '1B4', 0],
    'Reserved1': [(33, 84), '1I52', 0],
    'Spare1': [(85, 88), '1I4', 0],
    'Spare2': [(89, 92), '1I4', 0],
    'Reserved2': [(93, 124), '1I32', 0],
    'Spare3': [(125, 128), '1I4', 0],
    # PREFIX DATA PLATFORM REFERENCE INFORMATION
    'Platform information': [(129, 192), '1B64', 0],
    # PREFIX DATA - SENSOR/FACILITY SPECIFIC, AUXILIARY DATA
    'Fixed code = AA in Hexadecimal notation': [(193, 193), '1B1', 0],
    'OGRC/OBRC flag (1 or 0)': [(194, 194), '1B1', 0],
    'ICU on board time': [(195, 198), '1B4', 0],
    'Activity task': [(199, 200), '1B2', 0],
    'Image format counter': [(201, 204), '1B4', 0],
    'Sampling window start time': [(205, 206), '1B2', 0],
    'Pulse repetition interval': [(207, 208), '1B2', 0],
    'Calibration attenuation setting': [(209, 209), '1B1', 0],
    'Receiver gain attenuation setting': [(210, 210), '1B1', 0],
    'Spare4': [(211, 340), '130B1', 0],
    '36 calibration pulses as (4bit spare 6bit Q 6bit I from MSB down to LSB)': [(341, 412), '36B2', 0],
    # SAR RAW SIGNAL DATA
    'Raw Data': [(413, 414), '2B1', 0]  # [I, Q]
}

SarDataFileProcessedDataRecordERS = {
    'Record sequence number': [(1, 4), '1B4', 0],
    'l-st record sub-type code': [(5, 5), '1B1', 0],
    'Record type code': [(6, 6), '1B1', 0],
    '2-nd record sub-type code': [(7, 7), '1B1', 0],
    '3-rd record sub-type code': [(8, 8), '1B1', 0],
    'Length of this record': [(9, 12), '1B4', 0],
    # SAR PROCESSED DATA
    # 'Processed Data': [(13, 16), '2F2', 0],  # '2F2' for [real, imag], or '1C4' for real + 1j*imag
    'Processed Data': [(13, 16), '1I1', 0],  # '1I1' for amplitude image
}


def read_uavsar_csm(filepath, sl=1, el=-1, rmbp=False):
    r"""read UAVSAR raw data

    read UAVSAR SAR raw data pulse from line :attr:`sl` to line :attr:`el`. This function call
    function :func:`read_ceos_sar_raw` firstly and do some post-process.


    Parameters
    ----------
    filepath : str
        UAVSAR raw data file path string, for UAVSAR --> ``*.raw``
    sl : int, optional
        start line (azimuth) (the default is 1, which means the first line)
    el : int, optional
        end line (azimuth) (the default is -1, which means the last line)
    rmbp : bool, optional
        If you want to remove the padded border pixels, set :attr:`rmbp` to ``True``, else set to ``False`` (the default is ``False``).

    Returns
    -------
    S : 2d-array
       SAR raw signal data matrix.

    """

    logging.info("===In read_ers_sar_raw...")

    raise(ImportError('Not opened yet!'))


def read_uavsar_mlc(filepath, dshape=None, dtype='complex'):
    r"""read UAVSAR MLC data

    read UAVSAR SAR MLC data.


    Parameters
    ----------
    filepath : str
        UAVSAR MLC data file path string, for UAVSAR --> ``*.mlc``
    dshape : tuple, optional
        MLC data shape (:math:`N_a\times N_r`), where, :math:`N_a, N_r` equal
        to ``mlc_mag.set_rows``, ``mlc_mag.set_cols`` which can be obtained
        from the ``.ann`` file (the default is 1, which means the first line)
    dshape : str
        MLC data type: ``'complex'`` for complex-valued floating point data,
        ``'real'`` for real-valued floating point data.

    Returns
    -------
    S : 2d-array
       SAR MLC data matrix.

    """

    logging.info("===In read_uavsar_mlc...")

    raise(ImportError('Not opened yet!'))



#!/usr/bin/env python
#-*- coding: utf-8 -*-
# @file      : pulse_compression.py
# @author    : Zhi Liu
# @email     : zhiliu.mind@gmail.com
# @homepage  : http://iridescent.ink
# @date      : Sun Nov 27 2022
# @version   : 0.0
# @license   : The Apache License 2.0
# @note      : 
# 
# The Apache 2.0 License
# Copyright (C) 2013- Zhi Liu
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#

import numpy as np
import torch as th
from torchbox.base.arrayops import cut
from torchbox.utils.const import *


def mfpc_throwaway(X, No, Nh, axis=0, mffdmod='way1', ftshift=False):
    r"""Throwaway invalid pulse compressed data

    Throwaway invalid pulse compressed data

    Parameters
    ----------
    X : Tensor
        Data after pulse compression.
    No : int
        Output size.
    Nh : int
        Filter size.
    axis : int, optional
        Throwaway dimension. (the default is 0)
    mffdmod : str, optional
        Frequency filter mode. (the default is 'way1')
    ftshift : bool, optional
        Whether to shift frequency (the default is False)
    """

    if axis is None:
        axis = 1
    if ftshift is None:
        ftshift = False

    Nfft = np.size(X, axis)
    Nfft, No, Nh = np.uint([Nfft, No, Nh])

    if No > Nfft:
        raise ValueError('Output size is bigger than input size!')
    elif No == Nfft:
        return X

    N = Nfft - No

    if ftshift:
        if mffdmod in ['way1', 'WAY1', 'Way1', 'way2', 'WAY2', 'Way2', 'way3', 'WAY3', 'Way3', 'way4', 'WAY4', 'Way4']:
            Nstart = np.uint(np.floor((N + 1) / 2.))
            Nend = np.uint(Nfft - (N - Nstart))
            X = cut(X, ((Nstart, Nend),), axis)
        else:
            raise ValueError('Unsupported frequency domain matched filter: ' + mfmod)

    else:
        if mffdmod in ['way1', 'WAY1', 'Way1']:
            Nstart = np.uint(np.fix(Nh / 2.))
            Nend = np.uint(Nstart + No)
            X = cut(X, ((Nstart, Nend),), axis)
        elif mffdmod in ['way2', 'WAY2', 'Way2']:
            Nend1 = Nfft
            Nstart1 = np.uint(Nend1 - (Nh - 1))
            X = cut(X, ((Nstart1, Nend1), (0, No)), axis)
            Nstart = np.uint(np.fix(Nh / 2.))
            Nend = np.uint(Nstart + No)
            X = cut(X, ((Nstart, Nend),), axis=axis)
        elif mffdmod in ['way3', 'WAY3', 'Way3']:
            Nstart = 0
            Nend = No
            X = cut(X, ((Nstart, Nend),), axis)
        elif mffdmod in ['way4', 'WAY4', 'Way4']:
            Nstart = 0
            Nend = No
            X = cut(X, ((Nstart, Nend),), axis)
        else:
            raise ValueError('Unsupported frequency domain matched filter: ' + mfmod)
    return X


if __name__ == '__main__':

    X = th.tensor(range(32))
    print(X, X.shape[0])

    No, Nh = (24, 7)
    X = mfpc_throwaway(X, No, Nh, axis=0, mffdmod='way1', ftshift=True)
    print(X)



from __future__ import division, print_function, absolute_import

from .version import __version__
from .misc.sardata import SarData
from .misc.sarplat import SarPlat
from .misc.data import sarread, sarstore, format_data
from .misc.visual import saraxis, tgshow, apshow, show_response, \
    showReImAmplitudePhase, sarshow, show_image, show_sarimage, show_sarimage3d, imshow

from .sharing.scene_and_targets import gpts, gdisc, grectangle, dsm2tgs, tgs2dsm
from .sharing.compute_sar_parameters_func import compute_sar_parameters
from .sharing.chirp_signal import chirp_tran, chirp_recv, Chirp
from .sharing.sar_signal import sar_tran, sar_recv
from .sharing.matched_filter import chirp_mf_td, chirp_mf_fd
from .sharing.pulse_compression import mfpc_throwaway
from .sharing.range_migration import rcmc_interp
from .sharing.doppler_centroid_estimation import abdce_wda_opt, abdce_wda_ori, bdce_madsen, bdce_api, bdce_sf, fullfadc
from .sharing.doppler_rate_estimation import dre_geo
from .sharing.scatter_selection import center_dominant_scatters, window_data
from .sharing.antenna_pattern import antenna_pattern_azimuth
from .sharing.beamwidth_footprint import azimuth_beamwidth, azimuth_footprint
from .sharing.slant_ground_range import slantr2groundr, slantt2groundr, groundr2slantr, groundr2slantt, min_slant_range, min_slant_range_with_migration
from .sharing.sidelobe_suppression import sls_fd


from .sarcfg.sensor_satellite import SENSOR_SATELLITE
from .sarcfg.sensor_airbone import SENSOR_AIRBONE
from .sarcfg.sensor_simulated import SENSOR_SIMULATED

from .simulation.geometry import disc, rectangle
# from .simulation.chirp_scaling import sim_cs_tgs
from .simulation.simulation_time_domain import tgs2sar_td
from .simulation.simulation_freq_domain import dsm2sar_fd
from .simulation.sar_model import sarmodel
from .simulation.make_sar_params import SARParameterGenerator
from .simulation.make_targets import gpoints


from .autofocus.focusing import focus, defocus
from .autofocus.phase_error_model import convert_ppec, ppeaxis, polype, dctpe, rmlpe, PolyPhaseErrorGenerator
from .autofocus.phase_gradient import pgaf_sm
from .autofocus.minimum_entropy import meaf_ssa_sm, meaf_sm
from .autofocus.maximum_contrast import mcaf_sm
from .autofocus.fourier_domain_optim import af_ffo_sm

from .evaluation.target_background import extract_targets, tbr, tbr2


from .imaging.range_doppler_mftd import rda_mftd

from .sparse.sharing import sparse_degree
from .sparse.complex_image import CICISTA


from .calibration.channel_process import iq_correct
from .calibration.multilook_process import multilook_spatial
from .calibration.gain_compensation import vga_gain_compensation

from .module.autofocus.focusing import AutoFocus
from .module.autofocus.phase_error_model import BarePhi, PolyPhi, DctPhi, SinPhi
from .module.autofocus.fast_fourier_domain_optimization import AutoFocusFFO
from .module.autofocus.autofocus import AutoFocusBarePhi, AutoFocusPolyPhi, AutoFocusDctPhi, AutoFocusSinPhi

from .module.sharing.matched_filter import RangeMatchedFilter, AzimuthMatchedFilter, AzimuthMatchedFilterLinearFit
from .module.sharing.pulse_compression import RangeCompress, AzimuthCompress, AzimuthCompressLinearFit
from .module.sharing.range_migration import RangeMigrationCorrection

from .module.imaging.range_doppler_algorithm import LRDAnet

from .products.utils import getnumber, splitfmt
from .products.record import readrcd, readrcd1item, printrcd
from .products.ceos import decfmtfceos, read_ceos_sar_raw, read_ceos_sar_slc, SarDataFileFileDescriptorRecordCEOS, SarDataFileSignalDataRecordCEOS
from .products.ers import get_ers_sar_plat_position, read_ers_sar_ldr_iip, read_ers_sar_raw, read_ers_sar_slc, LeaderFileImportantImagingParametersRecordERS, SarDataFileFileDescriptorRecordERS, SarDataFileSignalDataRecordERS, SarDataFileProcessedDataRecordERS
from .products.radarsat import read_radarsat_sar_raw, SarDataFileFileDescriptorRecordRADARSAT, SarDataFileSignalDataRecordRADARSAT
from .products.uavsar import read_uavsar_csm, read_uavsar_mlc
from .products.alos import get_alos_palsar_plat_position, get_alos_palsar_attitude, read_alos_palsar_ldr_iip, read_alos_palsar_raw, read_alos_palsar_slc, LeaderFileImportantImagingParametersRecordALOS, SarDataFileFileDescriptorRecordALOSPALSAR, SarDataFileSignalDataRecordALOSPALSAR, SarImageFileFileDescriptorRecordALOSPALSAR, SarImageFileSignalDataRecordALOSPALSAR, SarImageFileProcessedDataRecordALOSPALSAR




#!/usr/bin/env python
#-*- coding: utf-8 -*-
# @file      : beamwidth_footprint.py
# @author    : Zhi Liu
# @email     : zhiliu.mind@gmail.com
# @homepage  : http://iridescent.ink
# @date      : Sun Nov 27 2022
# @version   : 0.0
# @license   : The Apache License 2.0
# @note      : 
# 
# The Apache 2.0 License
# Copyright (C) 2013- Zhi Liu
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#

from __future__ import division, print_function
import numpy as np
from torchbox.utils.const import *


def azimuth_beamwidth(Wl, La):
    BWa = 0.886 * Wl / La
    return BWa


def azimuth_footprint(R, Wl, La):

    BWa = 0.886 * Wl / La
    return R * BWa


def compute_range_beamwidth2(Nr, Fsr, H, Aon):
    r"""computes beam angle in range direction


    Parameters
    ----------
    Nr : int
        Number of samples in range direction.
    Fsr : float
        Sampling rate in range direction.
    H : float
        Height of the platform.
    Aon : float
        The off-nadir angle (unit: rad).

    Returns
    -------
    float
        The beam angle (unit, rad) in range direction.

    """

    Rnear = H / (np.cos(Aon) + EPS)
    Rfar = Rnear + Nr * (C / (2. * Fsr))

    return np.arccos(H / Rfar) - abs(Aon)


def cr_footprint(Wl, H, La, Ad):
    r"""cross range (azimuth) foot print

    .. math::
       R_{CR} \approx \frac{\lambda}{L_a}\frac{H}{{\rm cos}\theta_d}

    Parameters
    ----------
    Wl : float
        wave length
    H : float
        height of SAR platform
    La : float
        length of antenna aperture (azimuth)
    Ad : float
        depression angle

    Returns
    -------
    float
        foot print size in azimuth
    """

    FPa = (Wl * H) / (La * np.cos(Ad))

    return FPa



class LRDAnet(th.nn.Module):
    ...

    def __init__(self, Na, Nr, Fsa, Fsr, Ta, Tp, Ka, Kr, fdc, Fc, R0, Vr, La):
        ...

    def forward(self, X):
        ...

    def focus(self, X, pa=None, pr=None):




#!/usr/bin/env python
#-*- coding: utf-8 -*-
# @file      : range_migration.py
# @author    : Zhi Liu
# @email     : zhiliu.mind@gmail.com
# @homepage  : http://iridescent.ink
# @date      : Sun Nov 27 2022
# @version   : 0.0
# @license   : The Apache License 2.0
# @note      : 
# 
# The Apache 2.0 License
# Copyright (C) 2013- Zhi Liu
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#

import torch as th
import torchbox as tb


def rcmc_interp(Sr, tr, D):
    r"""range migration correction with linear interpolation

    Range migration correction with linear interpolation.

    Parameters
    ----------
    Sr : tensor
        SAR raw data :math:`N_a×N_r` in range dopplor domain
    tr : 1d-tensor
        time array :math:`N_r×1` in range
    D : 1d-tensor
        :math:`N_a×1` migration factor vector

    Returns
    -------
    Srrcmc
        data after range migration correction :math:`N_a×N_r`
    """

    if Sr.shape[1] != tr.shape[1]:
        raise ValueError('Sr has shape: ', Sr.shape,
                         'tr has shape: ', tr.shape)
    if Sr.shape[0] != D.shape[0]:
        raise ValueError('Sr has shape: ', Sr.shape,
                         'D has shape: ', D.shape)
    Sr = th.view_as_real(Sr)
    Sr = Sr.double()

    raise(ImportError('Not opened yet!'))




#!/usr/bin/env python
#-*- coding: utf-8 -*-
# @file      : ers.py
# @author    : Zhi Liu
# @email     : zhiliu.mind@gmail.com
# @homepage  : http://iridescent.ink
# @date      : Sun Nov 27 2022
# @version   : 0.0
# @license   : The Apache License 2.0
# @note      : 
# 
# The Apache 2.0 License
# Copyright (C) 2013- Zhi Liu
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#

import re
import copy
import logging
import numpy as np
from tqdm import tqdm
import torchsar as ts
import torchbox as tb
from torchsar.products.ceos import decfmtfceos
from torchsar.products.utils import splitfmt
from torchsar.products.record import readrcd, readrcd1item, printrcd
from torchsar.sarcfg.sensor_satellite import SENSOR_SATELLITE


LeaderFileImportantImagingParametersRecordERS = {
    'Radar frequency (GHz)': [(720 + 493, 720 + 500), '1F8', 0],
    'Radar wavelength (meters)': [(720 + 501, 720 + 516), '1F16', 0],

}


# B: number in binary format
# A: string
# I: char
SarDataFileFileDescriptorRecordERS = {
    # SAR DATA FILE - FILE DESCRIPTOR RECORD (FIXED SEGMENT)
    'Record sequence number': [(1, 4), '1B4', 0],

}

SarDataFileSignalDataRecordERS = {
    'Record sequence number': [(1, 4), '1B4', 0],

}

SarDataFileProcessedDataRecordERS = {
    'Record sequence number': [(1, 4), '1B4', 0],

}


def _getdtype_component(fmtrcd, fmtuser):

    raise(ImportError('Not opened yet!'))



def get_ers_sar_plat_position(D):
    """get platform position data

    return a numpy array
        1st: X Y Z Vx Vy Vz
        2st: X Y Z Vx Vy Vz
           :
        nst: X Y Z Vx Vy Vz

    Parameters
    ----------
    D : dict
        LeaderFileImportantImagingParametersRecordALOS dict
    """

    logging.info("===In get_ers_sar_platform_position...")

    raise(ImportError('Not opened yet!'))


def read_ers_sar_ldr_iip(filepath, verbose=False):
    r"""Read important imaging parameters from leader file

    Read important imaging parameters from leader file.

    Parameters
    ----------
    filepath : str
        Leader file path string.
    """

    logging.info("===In read_ers_sar_ldr_iip...")

    raise(ImportError('Not opened yet!'))


def read_ers_sar_raw(rawfile, ledfile, sl=1, el=-1, rmbp=False):
    r"""read ERS SAR raw data

    read ERS1/2 SAR raw data pulse from line :attr:`sl` to line :attr:`el`.


    Parameters
    ----------
    rawfile : str
        ERS SAR raw data file path string, for ERS1/2 --> ``*.raw``
        e.g. ``E2_81988_STD_L0_F327.000.raw``
    ledfile : str
        ERS SAR leader file path string, ---> ``*.ldr``, e.g. ``E2_81988_STD_L0_F327.000.ldr``
    sl : int, optional
        start line (azimuth) (the default is 1, which means the first line)
    el : int, optional
        end line (azimuth) (the default is -1, which means the last line)
    rmbp : bool, optional
        If you want to remove the padded border pixels, set :attr:`rmbp` to ``True``, else set to ``False`` (the default is ``False``).

    Returns
    -------
    S : 2d-array
       SAR raw signal data matrix.

    Pd : dict
       Parameter dictionary.

    """

    logging.info("===In read_ers_sar_raw...")

    raise(ImportError('Not opened yet!'))


def read_ers_sar_slc(filepath, sl=1, el=-1, rmbp=False):
    r"""read ERS SAR SLC data

    read ERS1/2 SAR SLC data pulse from line :attr:`sl` to line :attr:`el`.


    Parameters
    ----------
    filepath : str
        ERS SAR raw data file path string, for ERS1/2 --> ``*.raw``
    sl : int, optional
        start line (azimuth) (the default is 1, which means the first line)
    el : int, optional
        end line (azimuth) (the default is -1, which means the last line)
    rmbp : bool, optional
        If you want to remove the padded border pixels, set :attr:`rmbp` to ``True``, else set to ``False`` (the default is ``False``).

    Returns
    -------
    S : 2d-array
       SAR SLC data matrix.

    """

    logging.info("===In read_ers_sar_slc...")

    raise(ImportError('Not opened yet!'))




class SARParameterGenerator(object):
    """SAR Parameter Generator

    Attributes
    ----------
    prdict : dict
        ``{'p1': [low, high], 'p2': [low, high]...}``
    seed : int or None
        The seed for random generator.
    """

    def __init__(self, prdict, seed=None):
        ...

    def mksarp(self, n=1, seed=None):
        """Makes SAR Parameters

        Parameters
        ----------
        n : int, optional
            The number of experiments.
        seed : None, optional
            The seed for random generator.

        Returns
        -------
        dict
            The SAR Parameter dict.
        """





#!/usr/bin/env python
#-*- coding: utf-8 -*-
# @file      : matched_filter.py
# @author    : Zhi Liu
# @email     : zhiliu.mind@gmail.com
# @homepage  : http://iridescent.ink
# @date      : Sun Nov 27 2022
# @version   : 2.0
# @license   : The Apache License 2.0
# @note      : 
# 
# The Apache 2.0 License
# Copyright (C) 2013- Zhi Liu
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#

from __future__ import print_function
import torch as th
import torchbox as tb
from torch.nn.parameter import Parameter


class RangeMatchedFilter(th.nn.Module):

    def __init__(self, Na, Tp, Fsr, Kr, Fc, trainable=True, dtype=th.float64):
        super(RangeMatchedFilter, self).__init__()
        self.Na = Na
        self.Tp = Tp
        self.dtype = dtype  # filters with float64 is better then float32

        raise(ImportError('Not opened yet!'))

    def forward(self):

        raise(ImportError('Not opened yet!'))


class AzimuthMatchedFilter(th.nn.Module):

    def __init__(self, Nr, Tp, Fsa, Ka, Fc, trainable=True, dtype=th.float32):
        super(AzimuthMatchedFilter, self).__init__()
        self.Nr = Nr
        self.Tp = Tp
        self.dtype = dtype  # filters with float64 is better then float32

        raise(ImportError('Not opened yet!'))

    def forward(self):

        raise(ImportError('Not opened yet!'))


class AzimuthMatchedFilterLinearFit(th.nn.Module):

    def __init__(self, Nr, Tp, Fsa, Ka, Fc, trainable=True, dtype=th.float32):
        super(AzimuthMatchedFilterLinearFit, self).__init__()
        self.Nr = Nr
        self.Tp = Tp
        self.dtype = dtype  # filters with float64 is better then float32

        if type(trainable) is bool:
            trainable = [trainable] * 2

        raise(ImportError('Not opened yet!'))

    def forward(self):

        raise(ImportError('Not opened yet!'))


if __name__ == '__main__':

    pass



def center_dominant_scatters(x, axis=-2, isplot=False):
    """Center the dominant scatters.

    Parameters
    ----------
    x : tensor
        Complex SAR image tensor.
    axis : int, optional
        The axis to be centered in.
    isplot : bool, optional
        Whether to plot.

    Returns
    -------
    tensor
        Centered.

    Raises
    ------
    TypeError
        Intput must be complex image!
    """

def window_data(z, win=None, axis=-2, isplot=False):




def __entropy(X):
    ...

def meaf_ssa_sm(x, niter=10, delta=None, toli=1e-2, tolo=1e-2, ftshift=True, islog=False):
    r"""Stage-by-Stage minimum entropy

    [1] Morrison Jr, Robert Lee; Autofocus, Entropy-based (2002): Entropy-based autofocus for synthetic aperture radar.

    Parameters
    ----------
    x : Tensor
       Corrupt complex SAR image. N-Na-Nr(complex) or N-Na-Nr-2(real)
    niter : int, optional
       The number of iteration (the default is 10)
    delta : {float or None}, optional
       The change step (the default is None (i.e. PI))
    toli : int, optional
       Tolerance error for inner loop (the default is 1e-2)
    tolo : int, optional
       Tolerance error for outer loop (the default is 1e-2)
    ftshift : bool, optional
       Shift the zero frequency to center? (the default is True)
    islog : bool, optional
       Print log information? (the default is False)
    """

def meaf_sm(x, phi=None, niter=10, tol=1e-4, eta=0.1, method='N-MEA', selscat=False, isunwrap=True, deg=None, axis=-2, ftshift=True, islog=False):
    r"""Entropy based autofocus

    Minimum-Entropy based Autofocus (MEA)

    Args:
        x (Tensor): complex image with shape :math:`N×N_a×N_r` or :math:`N×N_a×N_r×2`
        phi (Tensor, optional): initial value of :math:`\phi` (the default is None, which means zeros)
        niter (int, optional): number of iterations (the default is 10)
        tol (float, optional): Error tolerance.
        eta (float, optional): Learning rate.
        method (str, optional): method used to update the phase error
            - ``'FP-MEA'`` --> Fix Point
            - ``'CD-MEA'`` --> Coordinate Descent
            - ``'SU-MEA'`` --> Simultaneous Update
            - ``'N-MEA'`` --> Newton, see [2], [1] has problem when used to small image
            - ``'SN-MEA'`` --> Simplified Newton
            - ``'MN-MEA'`` --> Modified Newton
        selscat (bool, optional): Select brighter scatters (default: False).
        isunwrap (bool, optional): Unwrap radian phase (default: True).
        deg (int or None, optional): The degree of polynominal, if None, no fitting.
        axis (int, optional): Compensation axis.
        ftshift (bool, optional): Does shift zero frequency to center?
        islog (bool, optional): Does print log info.


        see [1] Zhang S , Liu Y , Li X . Fast Entropy Minimization Based Autofocusing Technique for ISAR Imaging[J]. IEEE Transactions on Signal Processing, 2015, 63(13):3425-3434.
            [2] Zeng T , Wang R , Li F . SAR Image Autofocus Utilizing Minimum-Entropy Criterion[J]. IEEE Geoence & Remote Sensing Letters, 2013, 10(6):1552-1556.

    Returns:
        (Tensor): Description
    """





#!/usr/bin/env python
#-*- coding: utf-8 -*-
# @file      : visual.py
# @author    : Zhi Liu
# @email     : zhiliu.mind@gmail.com
# @homepage  : http://iridescent.ink
# @date      : Sun Nov 27 2022
# @version   : 0.0
# @license   : The Apache License 2.0
# @note      : 
# 
# The Apache 2.0 License
# Copyright (C) 2013- Zhi Liu
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#

import torch as th
import numpy as np
import matplotlib.pyplot as plt
from torchbox.utils.const import *
from torchbox.base.mathops import fnab
from torchsar.sharing.scene_and_targets import tgs2dsm
from torchsar.sharing.slant_ground_range import slantr2groundr, groundr2slantr
from skimage import exposure


def saraxis(pdict, axismod=None):
    r"""generate SAR image axes ticks and strings

    Parameters
    ----------
    pdict : dict
        SAR parameter dictionary.
    axismod : str, optional
        axis mode, supported are ``'Image'`` ``'SceneAbsoluteSlantRange'``, ``'SceneRelativeSlantRange'``,
        ``'SceneAbsoluteGroundRange'``, ``'SceneRelativeGroundRange'``, ``'BeamAreaSceneArea'``
         by default None (``'Image'``)

    Returns
    -------
    list and str
        extent, xlabel string, ylabel string, title string.
    """    

    GM = pdict['GeometryMode']
    H = pdict['H']
    V = pdict['V']
    ta = pdict['ta']
    tr = pdict['tr']
    Ar = pdict['Ar']
    if 'SceneArea' in pdict.keys():
        SceneArea = pdict['SceneArea']
    if 'SubSceneArea' in pdict.keys():
        SceneArea = pdict['SubSceneArea']
    if 'BeamArea' in pdict.keys():
        BeamArea = pdict['BeamArea']
    if 'SubBeamArea' in pdict.keys():
        BeamArea = pdict['SubBeamArea']
    if 'SceneCenter' in pdict.keys():
        SceneCenter = pdict['SceneCenter']
    if 'SubSceneCenter' in pdict.keys():
        SceneCenter = pdict['SubSceneCenter']
    if 'BeamCenter' in pdict.keys():
        BeamCenter = pdict['BeamCenter']
    if 'SubBeamCenter' in pdict.keys():
        BeamCenter = pdict['SubBeamCenter']
    if 'EchoSize' in pdict.keys():
        Na, Nr = pdict['EchoSize'][0:2]
    if 'SubEchoSize' in pdict.keys():
        Na, Nr = pdict['SubEchoSize'][0:2]
    if 'ES' in pdict.keys():
        Na, Nr = pdict['ES'][0:2]

    nTicks = 10

    if axismod is None:
        axismod = 'Image'

    if axismod == 'Image':
        extent = [1, Nr, 1, Na]
        xlabelstr = "Range"
        ylabelstr = "Azimuth"
    elif axismod == 'SceneAbsoluteSlantRange':
        if GM in ['BeamGeometry', 'BG']:
            extent = [pdict['Rnear'], pdict['Rfar'], BeamArea[2], BeamArea[3]]
        if GM in ['SceneGeometry', 'SG']:
            extent = [pdict['Rnear'], pdict['Rfar'], SceneArea[2], SceneArea[3]]
        xlabelstr = "Range (m)"
        ylabelstr = "Azimuth (m)"
    elif axismod == 'SceneRelativeSlantRange':
        if GM in ['BeamGeometry', 'BG']:
            extent = [pdict['Rnear'] - pdict['Rbc'], pdict['Rfar'] - pdict['Rbc'], BeamArea[2], BeamArea[3]]
        if GM in ['SceneGeometry', 'SG']:
            extent = [pdict['Rnear'] - pdict['Rsc'], pdict['Rfar'] - pdict['Rsc'], SceneArea[2], SceneArea[3]]
        xlabelstr = "Range (m)"
        ylabelstr = "Azimuth (m)"
    elif axismod == 'SceneAbsoluteGroundRange':
        if GM in ['BeamGeometry', 'BG']:
            X = slantr2groundr(C * tr / 2., H, Ar, 0)
            extent = [X[0], X[-1], BeamArea[2], BeamArea[3]]
        if GM in ['SceneGeometry', 'SG']:
            X = slantr2groundr(C * tr / 2., H, Ar, 0)
            extent = [X[0], X[-1], SceneArea[2], SceneArea[3]]
        # xticks = X[::int(len(X) / nTicks)]
        xlabelstr = "Range (m)"
        ylabelstr = "Azimuth (m)"
    elif axismod == 'SceneRelativeGroundRange':
        if GM in ['BeamGeometry', 'BG']:
            X = slantr2groundr(C * tr / 2., H, Ar, BeamCenter[0])
            extent = [X[0, 0], X[0, -1], BeamArea[2], BeamArea[3]]
        if GM in ['SceneGeometry', 'SG']:
            X = slantr2groundr(C * tr / 2., H, Ar, SceneCenter[0])
            extent = [X[0, 0], X[0, -1], SceneArea[2], SceneArea[3]]
        # xticks = X[::int(len(X) / nTicks)]
        xlabelstr = "Range (m)"
        ylabelstr = "Azimuth (m)"
    elif axismod == 'BeamAreaSceneArea':
        if GM in ['BeamGeometry', 'BG']:
            extent = BeamArea
        if GM in ['SceneGeometry', 'SG']:
            extent = SceneArea
        xlabelstr = "Range (m)"
        ylabelstr = "Azimuth (m)"
    else:
        extent = [ta[0, 0] * V, ta[-1, 0] * V, tr[0, 0] * C, tr[-1, 0] * C]
        xlabelstr = 'x'
        ylabelstr = 'y'
    extent = np.array(extent)
    return extent, xlabelstr, ylabelstr


def tgshow(targets, scene, shape, extent=None, cmap=None, isflip=[False, False], labelstr=None, outfile=None, isshow=True):
    r"""show targets

    Show targets in an image.

    Arguments
    --------------
    targets : Tensor
        A tensor contains information of targets.
    scene : tuple or list
        Area of scene [xmin, xmax, ymin, ymax].
    shape : tuple or list
        The shape of the scene.

    Keyword Arguments
    --------------
    outfile : str
        The filename for writting figure (default: None, do not save).
    isshow : bool
        Whether to plot figure (default: True).

    Returns
    --------------
        Tensor -- The final image tensor for show.

    Raises
    --------------
        ValueError -- :attr:`targets` should not be None.
    """

    if targets is None:
        raise ValueError("targets should not be None")

    II = tgs2dsm(targets, scene=scene, bg=0, dsize=shape, device='cpu')

    if isflip[0]:
        II = np.flipud(II)
    if isflip[1]:
        II = np.fliplr(II)

    plt.figure()
    if extent is None:
        plt.imshow(II, cmap=cmap, aspect=None, interpolation='none')
    else:
        plt.imshow(II, extent=extent, cmap=cmap, aspect=None, interpolation='none')

    if labelstr is not None:
        if labelstr[0] is not None:
            plt.xlabel(labelstr[0])
        if labelstr[1] is not None:
            plt.ylabel(labelstr[1])
        if labelstr[-1] is not None:
            plt.title(labelstr[-1])
    plt.colorbar()

    if outfile is not None:
        plt.savefig(outfile)
        print("target image has been saved to: ", outfile)
    if isshow:
        plt.show()
    plt.close()
    return II


def apshow(Srx, Title=None, cmap=None, extent=None, keepAspectRatio=True, outfile=None, isshow=True):
    r"""Show amplitude and phase.

    Show amplitude and phase.

    Parameters
    ----------
    Srx : Tensor
        Complex-valued tensor.
    Title : str, optional
        The figure title (the default is None, which means
        no title).
    cmap : str, optional
        The colormap (the default is None)
    extent : tuple or list, optional
        [description] (the default is None)
    keepAspectRatio : bool, optional
        [description] (the default is True)
    outfile : str, optional
        The filename for outputing (the default is None, which means not save)
    isshow : bool, optional
        Whether to show the figure (the default is True).
    """
    if Title is None:
        Title = 'SAR raw data'

    plt.figure()
    plt.subplot(121)
    plt.imshow(np.absolute(
        # Srx), aspect='auto' if not keepAspectRatio else None,
        # interpolation='none')
        Srx), extent=extent, aspect='auto' if not keepAspectRatio else None, interpolation='none', cmap=cmap)
    # plt.xlabel("Range (m)")
    # plt.ylabel("Azimuth (m)")
    plt.xlabel("Range\n(a)")
    plt.ylabel("Azimuth")
    plt.title(Title + " (amplitude)")

    plt.subplot(122)
    # plt.xlabel("Range (m)")
    # plt.ylabel("Azimuth (m)")
    plt.xlabel("Range\n(b)")
    plt.ylabel("Azimuth")
    plt.imshow(np.angle(Srx), extent=extent,
               aspect='auto' if not keepAspectRatio else None, interpolation='none', cmap=cmap)
    plt.title(Title + " (phase)")
    plt.tight_layout()
    if outfile is not None:
        plt.savefig(outfile)
        print("image has been saved to: ", outfile)
    # plt.colorbar()
    if isshow:
        plt.show()
    else:
        plt.close()


def show_response(Srx, extent=None, title="Figure", keepAspectRatio=True, outfile=None, isshow=True):
    r"""show SAR response

    [description]

    Arguments
    ------------------
    Srx : Tensor
        [description]

    Keyword Arguments
    ------------------
    extent : tuple, list or None
        [description] (default: None)
    title : str
        [description] (default: "Figure")
    keepAspectRatio : bool
        [description] (default: True)
    outfile : str
        [description] (default: None)
    isshow : bool
        [description] (default: True)
    """

    plt.figure()
    plt.title(title)
    plt.xlabel("Range (m)")
    plt.ylabel("Azimuth (m)")
    extent = [27844.409098089047, 28868.409098089047, 512.0, -512.0]
    plt.imshow(np.absolute(Srx), extent=extent,
               aspect='auto' if not keepAspectRatio else None, interpolation='none')
    plt.colorbar()
    if outfile is not None:
        plt.savefig(outfile)
        print("image has been saved to: ", outfile)
    if isshow:
        plt.show()
    else:
        plt.close()


def showReImAmplitudePhase(Srx, extent=None, title=None, keepAspectRatio=True):
    r"""[summary]

    [description]

    Parameters
    ----------
    Srx : {[type]}
        [description]
    extent : {[type]}, optional
        [description] (the default is None, which [default_description])
    title : str, optional
        [description] (the default is None, which [default_description])
    keepAspectRatio : bool, optional
        [description] (the default is True, which [default_description])
    """

    f = plt.figure()
    f.suptitle(title)
    plt.subplot(221)
    plt.title("SAR response, module")
    # plt.xlabel("Range (m)")
    # plt.ylabel("Azimuth (m)")
    plt.xlabel("Range")
    plt.ylabel("Azimuth")
    plt.imshow(np.absolute(Srx), extent=extent,
               aspect='auto' if not keepAspectRatio else None, interpolation='none')
    plt.colorbar()
    plt.subplot(222)
    plt.title("SAR response (phase)")
    plt.xlabel("Range")
    plt.ylabel("Azimuth")
    plt.imshow(np.angle(Srx), extent=extent,
               aspect='auto' if not keepAspectRatio else None, interpolation='none')
    plt.colorbar()
    plt.subplot(223)
    plt.title("SAR response (real part)")
    plt.xlabel("Range")
    plt.ylabel("Azimuth")
    plt.imshow(np.real(Srx), extent=extent,
               aspect='auto' if not keepAspectRatio else None, interpolation='none')
    plt.colorbar()
    plt.subplot(224)
    plt.title("SAR response (imaginary part)")
    plt.xlabel("Range")
    plt.ylabel("Azimuth")
    plt.imshow(np.imag(Srx), extent=extent,
               aspect='auto' if not keepAspectRatio else None, interpolation='none')
    plt.colorbar()
    plt.show()


def show_image(img, Title=None, cmap=None, keepAspectRatio=True, outfile=None, isshow=True):
    H = img.shape[0]
    W = img.shape[1]

    if Title is None:
        # Title = "Image"
        Title = "Intensity"
    extent = [-W / 2.0, W / 2.0, -H / 2.0, H / 2.0]
    # extent = [0, W, H, 0]

    plt.figure()
    ax = plt.subplot(111)
    plt.title(Title)
    plt.xlabel("Range")
    plt.ylabel("Azimuth")
    # plt.xticks(fontsize=17)
    # plt.yticks(fontsize=17)
    # ax.set_xlabel("Range", fontsize=17)
    # ax.set_ylabel("Azimuth", fontsize=17)
    # ax.legend(fontsize=17)
    plt.imshow(img, extent=extent,
               aspect='auto' if not keepAspectRatio else None, interpolation='none', cmap=cmap)

    plt.colorbar()

    if outfile is not None:
        plt.savefig(outfile)
        print("image has been saved to: ", outfile)

    if isshow:
        plt.show()
        pass
    else:
        plt.close()
    return img


def sarshow(Xs, pdict, axismod=None, nrows=None, ncols=None, xlabels=None, ylabels=None, titles=None, figsize=None, outfile=None, **kwargs):
    """_summary_

    Parameters
    ----------
    Xs : tensor, array, list or tuple
        list/tuple of image arrays/tensors, if the type is not list or tuple, wrap it.
    pdict : dict
        SAR parameter dict
    axismod : _type_, optional
        _description_, by default None
    nrows : int, optional
        show in :attr:`nrows` rows, by default None (auto computed).
    ncols : int, optional
        show in :attr:`ncols` columns, by default None (auto computed).
    xlabels : str, optional
        labels of x-axis
    ylabels : str, optional
        labels of y-axis
    titles : str, optional
        titles
    figsize : tuple, optional
        figure size, by default None
    outfile : str, optional
        save image to file, by default None (do not save).
    kwargs : 
        see :func:`matplotlib.pyplot.imshow`
    """

    if (type(Xs) is not list) and (type(Xs) is not tuple):
        Xs = [Xs]

    extent, xlabelstr, ylabelstr = saraxis(pdict=pdict, axismod=axismod)

    n = len(Xs)
    nrows, ncols = fnab(n)
    xlabels = [xlabels] * n if (type(xlabels) is str) or (xlabels is None) else xlabels
    ylabels = [ylabels] * n if (type(ylabels) is str) or (ylabels is None) else ylabels
    titles = [titles] * n if (type(titles) is str) or (titles is None) else titles
    plt.figure(figsize=figsize)
    for i, X, xlabel, ylabel, title in zip(range(n), Xs, xlabels, ylabels, titles):
        plt.subplot(nrows, ncols, i + 1)
        if type(X) is th.Tensor:
            X = X.cpu().numpy()
        if axismod in ['image', 'Image', 'IMAGE']:
            plt.imshow(X, **kwargs)
            plt.xlabel(xlabel)
            plt.ylabel(ylabel)
            plt.title(title)
        else:
            plt.imshow(X, extent=extent, **kwargs)
            plt.xlabel(xlabelstr)
            plt.ylabel(ylabelstr)
            plt.title(title)

    if outfile is not None:
        plt.savefig(outfile)
        print("sar image has been saved to: ", outfile)

    return plt


def show_sarimage(SI, pdict, axismod=None, title=None, cmap=None, isimgadj=False, aspect=None, outfile=None, newfig=True, figsize=None):
    r"""[summary]

    [description]

    Arguments:
        SI {[type]} -- [description]
        pdict {[type]} -- [description]

    Keyword Arguments:
        axismod {[type]} -- [description] (default: {None})
        title {[type]} -- [description] (default: {None})
        cmap {[type]} -- [description] (default: {None})
        isimgadj bool -- [description] (default: {False})
        aspect {[type]} -- [description] (default: {None})
        outfile {[type]} -- [description] (default: {None})
        newfig bool -- [description] (default: {True})
        figsize {[type]} -- [description] (default: {None})
    """

    extent, xlabelstr, ylabelstr = saraxis(pdict=pdict, axismod=axismod)

    if newfig:
        plt.figure(figsize=figsize)
        ax = plt.subplot(111)

    plt.xlabel(xlabelstr)
    plt.ylabel(ylabelstr)

    SI = np.flipud(SI)

    A = np.absolute(SI)

    plt.title(title)

    if isimgadj:
        A = A / A.max()
        A = 20 * np.log10(A + EPS)
        a = np.abs(A.mean())
        A = (A + a) / a * 255
        A[A < 0] = 0
        A = A.astype('uint8')
        A = np.flipud(A)

    if axismod in ['image', 'Image', 'IMAGE']:
        ax.imshow(A, aspect=aspect, interpolation='none', cmap=cmap)
    else:
        ax.imshow(A, extent=extent, aspect=aspect, interpolation='none', cmap=cmap)

    if axismod == 'SceneAbsoluteGroundRange' or axismod == 'SceneRelativeGroundRange':
        # plt.xticks(xticks)
        pass
        # ax.xaxis.set_major_locator(FixedLocator(X[::int(len(X) / 10.)]))
    # plt.colorbar()
    plt.tight_layout()
    if outfile is not None:
        plt.savefig(outfile)
        print("sar image has been saved to: ", outfile)

    if newfig:
        plt.show()


def show_sarimage3d(SI, pdict, axismod=None, title=None, cmap=None, isimgadj=False, aspect=None, outfile=None, figsize=None):
    r"""[summary]

    [description]

    Arguments:
        SI {[type]} -- [description]
        sarplat {[type]} -- [description]

    Keyword Arguments:
        axismod {[type]} -- [description] (default: {None})
        title {[type]} -- [description] (default: {None})
        cmap {[type]} -- [description] (default: {None})
        isimgadj bool -- [description] (default: {False})
        aspect {[type]} -- [description] (default: {None})
        outfile {[type]} -- [description] (default: {None})
        figsize {[type]} -- [description] (default: {None})
    """

    extent, xlabelstr, ylabelstr = saraxis(pdict=pdict, axismod=axismod)

    Z = np.absolute(SI)
    if isimgadj:
        Z = exposure.adjust_log(Z)
    Z = np.flipud(Z)
    # fig = plt.figure(figsize=figsize)
    M, N = np.shape(Z)

    X = np.arange(0, N, 1)
    Y = np.arange(0, M, 1)
    X, Y = np.meshgrid(X, Y, indexing='xy')

    # mlab.figure(size=(400, 500))
    # mlab.mesh(X, Y, Z)
    # mlab.surf(X, Y, Z)
    # mlab.colorbar()
    # mlab.xlabel(xlabelstr)
    # mlab.ylabel(ylabelstr)
    # mlab.zlabel("Amplitude")
    # mlab.title(title)
    # mlab.show()

    if outfile is not None:
        # mlab.savefig(outfile)
        print("sar image has been saved to: ", outfile)


def imshow(Xs, nrows=None, ncols=None, xlabels=None, ylabels=None, titles=None, figsize=None, outfile=None, **kwargs):
    r"""show images

    This function create an figure and show images in :math:`a` rows and :math:`b` columns.

    Parameters
    ----------
    Xs : tensor, array, list or tuple
        list/tuple of image arrays/tensors, if the type is not list or tuple, wrap it.
    nrows : int, optional
        show in :attr:`nrows` rows, by default None (auto computed).
    ncols : int, optional
        show in :attr:`ncols` columns, by default None (auto computed).
    xlabels : str, optional
        labels of x-axis
    ylabels : str, optional
        labels of y-axis
    titles : str, optional
        titles
    figsize : tuple, optional
        figure size, by default None
    outfile : str, optional
        save image to file, by default None (do not save).
    kwargs : 
        see :func:`matplotlib.pyplot.imshow`

    Returns
    -------
    plt
        plot handle
    """

    if (type(Xs) is not list) and (type(Xs) is not tuple):
        Xs = [Xs]

    n = len(Xs)
    nrows, ncols = fnab(n)
    xlabels = [xlabels] * n if (type(xlabels) is str) or (xlabels is None) else xlabels
    ylabels = [ylabels] * n if (type(ylabels) is str) or (ylabels is None) else ylabels
    titles = [titles] * n if (type(titles) is str) or (titles is None) else titles
    plt.figure(figsize=figsize)
    for i, X, xlabel, ylabel, title in zip(range(n), Xs, xlabels, ylabels, titles):
        plt.subplot(nrows, ncols, i + 1)
        if type(X) is th.Tensor:
            X = X.cpu().numpy()
        plt.imshow(X, **kwargs)
        plt.xlabel(xlabel)
        plt.ylabel(ylabel)
        plt.title(title)

    if outfile is not None:
        plt.savefig(outfile)
        
    return plt


if __name__ == '__main__':

    x = th.rand(5, 100, 100)
    plt = imshow([xi for xi in x], xlabels='x', ylabels='y', titles='title', cmap='hsv')
    plt.show()



def chirp_mf_td(K, Tp, Fs, Fc=0., Ns=None, mod='conv', scale=False, device='cpu'):
    """Generates matched filter of chirp signal in time domain

    Generates matched filter of chirp signal in time domain.

    Parameters
    ----------
    K : int
        The chirp rate.
    Tp : float
        The pulse width.
    Fs : float
        The sampling rate.
    Fc : float, optional
        The center frequency.
    Ns : int or None, optional
        The number of samples.
    mod : str, optional
        The mode of filter, ``'conv'`` or ``'corr'``
    scale : bool, optional
        Whether to scale the amplitude of the filter.
    device : str, optional
        Specifies the device to be used for computing.

    Returns
    -------
    tensor
        The matched filter tensor.

    Examples
    --------

    Do Pulse Compression with time/frequency domain filters.

    .. image:: ./_static/DemoPulseCompression.png
       :scale: 100 %
       :align: center

    The results shown in the above figure can be obtained by the following codes.

    ::

        #!/usr/bin/env python
        # -*- coding: utf-8 -*-
        # @Date    : 2019-02-18 10:14:12
        # @Author  : Yan Liu & Zhi Liu (zhiliu.mind@gmail.com)
        # @Link    : http://iridescent.ink
        # @Version : $1.0$
        import torch as th
        import torchbox as tb
        import torchsar as ts
        import matplotlib.pyplot as plt

        from torchbox.utils.const import *
        from torchbox.dsp.ffts import fft, ifft, fftfreq

        mftdmod = 'conv'
        mffdmod = 'way1'
        ftshift = True

        # ===Generate tansmit= ted and recieved signals
        # ---Setting parameters
        R = [1.e3, 2.e3, 3.e3]
        G = [0.5, 1.0, 0.8]

        EPS = 2.2e-32
        K = 4.1e+11
        Tp = 37.0e-06
        B = abs(K) * Tp

        alp = 1.24588  # 1.1-1.4
        Fs = alp * B
        Fc = 5.3e9
        Fc = 0.

        Koff = 0.
        Koff = 1.e10
        Fcoff = 0.
        # Fcoff = 1.e8

        Ts = 2.1 * Tp
        Ns = round(Fs * Ts)
        Nh = round(Fs * Tp)
        t = th.linspace(-Ts / 2., Ts / 2, Ns)
        f = fftfreq(Ns, Fs, norm=False, shift=ftshift)

        N = Ns + Nh - 1
        Nfft = 2**tb.nextpow2(N)

        # ---Transmitted signal
        St = ts.chirp_tran(t, Tp, K, Fc, a=1.)

        # ---Recieved signal
        Sr = ts.chirp_recv(t, Tp, K, Fc, a=1., g=G, r=R)

        # ---
        chirp = ts.Chirp(Tp=Tp, K=K, Fc=Fc, a=1.)

        St = chirp.tran(t)
        Sr = chirp.recv(t, g=G, r=R)


        # ---Frequency domain
        Yt = fft(St, dim=0, shift=ftshift)
        Yr = fft(Sr, dim=0, shift=ftshift)

        fontsize = 12
        fonttype = 'Times New Roman'
        fontdict = {'family': fonttype, 'size': fontsize}

        # ---Plot signals
        plt.figure(figsize=(10, 8))
        plt.subplot(221)
        plt.plot(t * 1e6, th.real(St))
        plt.grid()
        plt.title('Real part', fontdict=fontdict)
        plt.xlabel(r'Time/$\mu s$', fontdict=fontdict)
        plt.ylabel('Amplitude', fontdict=fontdict)
        plt.xticks(fontproperties=fonttype, size=fontsize)
        plt.yticks(fontproperties=fonttype, size=fontsize)
        plt.subplot(222)
        plt.plot(t * 1e6, th.imag(St))
        plt.grid()
        plt.title('Imaginary part', fontdict=fontdict)
        plt.xlabel(r'Time/$\mu s$', fontdict=fontdict)
        plt.ylabel('Amplitude', fontdict=fontdict)
        plt.xticks(fontproperties=fonttype, size=fontsize)
        plt.yticks(fontproperties=fonttype, size=fontsize)
        plt.subplot(223)
        plt.plot(f, th.abs(Yt))
        plt.grid()
        plt.title('Spectrum', fontdict=fontdict)
        plt.xlabel('Frequency/Hz', fontdict=fontdict)
        plt.ylabel('Amplitude', fontdict=fontdict)
        plt.xticks(fontproperties=fonttype, size=fontsize)
        plt.yticks(fontproperties=fonttype, size=fontsize)
        plt.subplot(224)
        plt.plot(f, th.angle(Yt))
        plt.grid()
        plt.title('Spectrum', fontdict=fontdict)
        plt.xlabel('Frequency/Hz', fontdict=fontdict)
        plt.ylabel('Phase', fontdict=fontdict)
        plt.xticks(fontproperties=fonttype, size=fontsize)
        plt.yticks(fontproperties=fonttype, size=fontsize)
        plt.subplots_adjust(left=0.08, bottom=0.06, right=0.98, top=0.96, wspace=0.19, hspace=0.25)

        plt.show()

        print(th.sum(t), th.sum(St), th.sum(f))


        plt.figure(figsize=(10, 8))
        plt.subplot(221)
        plt.plot(t * 1e6, th.real(Sr))
        plt.grid()
        plt.title('Real part', fontdict=fontdict)
        plt.xlabel(r'Time/$\mu s$', fontdict=fontdict)
        plt.ylabel('Amplitude', fontdict=fontdict)
        plt.subplot(222)
        plt.plot(t * 1e6, th.imag(Sr))
        plt.grid()
        plt.title('Imaginary part', fontdict=fontdict)
        plt.xlabel(r'Time/$\mu s$', fontdict=fontdict)
        plt.ylabel('Amplitude', fontdict=fontdict)
        plt.subplot(223)
        plt.plot(f, th.abs(Yr))
        plt.grid()
        plt.title('Spectrum', fontdict=fontdict)
        plt.xlabel('Frequency/Hz', fontdict=fontdict)
        plt.ylabel('Amplitude', fontdict=fontdict)
        plt.subplot(224)
        plt.plot(f, th.angle(Yr))
        plt.grid()
        plt.title('Spectrum', fontdict=fontdict)
        plt.xlabel('Frequency/Hz', fontdict=fontdict)
        plt.ylabel('Phase/rad', fontdict=fontdict)
        plt.subplots_adjust(left=0.08, bottom=0.06, right=0.98, top=0.96, wspace=0.19, hspace=0.25)

        plt.show()


        # ===Matched filtering/Pulse compression in time domain

        # ---Matched filtering in time domain
        Sm, tm = ts.chirp_mf_td(K + Koff, Tp, Fs, Fc=Fc + Fcoff, Ns=None, mod=mftdmod)

        if mftdmod in ['conv', 'Conv']:
            # S1 = th.convolve(Sr, Sm, mode='same')
            # S1 = ts.conv1(Sr, Sm, shape='same')
            S1 = tb.fftconv1(Sr, Sm, shape='same')

        if mftdmod in ['corr', 'Corr']:
            # S1 = th.correlate(Sr, Sm, mode='same')
            # S1 = ts.corr1(Sr, Sm, shape='same')
            S1 = tb.fftcorr1(Sr, Sm, shape='same')

        # ===Matched filtering/Pulse compression in frequency domain

        # ---Matched filter in frequency domain
        # ~~~Method 1-4
        H, f = ts.chirp_mf_fd(K + Koff, Tp, Fs, Fc=Fc + Fcoff, Nfft=Nfft, mod=mffdmod, win=None, ftshift=ftshift)

        print(H.shape, t.shape, Sm.shape, "=======")

        # ---Tansform the recieved signal to frequency domain
        Yr = fft(tb.padfft(Sr, Nfft, 0, ftshift), Nfft, dim=0, shift=ftshift)

        # ---Matched filtering/Pulse compression
        Y = Yr * H

        # ---Tansform back to time domain
        S2 = ifft(Y, Nfft, dim=0, shift=ftshift)

        S2 = ts.mfpc_throwaway(S2, Ns, Nh, 0, mffdmod, ftshift)


        print(th.sum(th.abs(S1 - S2)))
        print(th.sum(th.abs(S1 - S2)))
        print(th.argmax(th.abs(S1)), th.argmax(th.abs(S2)))


        plt.figure(figsize=(12, 8))
        plt.subplot(421)
        plt.plot(t * 1e6, th.real(St), '-', linewidth=1)
        plt.plot(t * 1e6, th.abs(St), '--', linewidth=2)
        plt.grid()
        plt.legend(('Real part', 'Amplitude'))
        plt.title('Transmitted', fontdict=fontdict)
        plt.xlabel(r'Time/$\mu s$', fontdict=fontdict)
        plt.ylabel('Amplitude', fontdict=fontdict)
        plt.subplot(422)
        plt.plot(t * 1e6, th.real(Sr), '-', linewidth=1)
        plt.plot(t * 1e6, th.abs(Sr), '--', linewidth=2)
        plt.grid()
        plt.legend(('Real part', 'Amplitude'))
        plt.title('Recieved', fontdict=fontdict)
        plt.xlabel(r'Time/$\mu s$', fontdict=fontdict)
        plt.ylabel('Amplitude', fontdict=fontdict)
        plt.subplot(423)
        plt.plot(tm * 1e6, th.real(Sm), '-', linewidth=1)
        plt.plot(tm * 1e6, th.abs(Sm), '--', linewidth=2)
        plt.grid()
        plt.legend(('Real part', 'Amplitude'))
        plt.title('Filter (time domain)', fontdict=fontdict)
        plt.xlabel(r'Time/$\mu s$', fontdict=fontdict)
        plt.ylabel('Amplitude', fontdict=fontdict)
        plt.subplot(424)
        plt.plot(f, th.real(H), '-', linewidth=1)
        plt.plot(f, th.abs(H), '--', linewidth=2)
        plt.grid()
        plt.legend(('Real part', 'Amplitude'))
        plt.title('Filter (frequency domain,' + mffdmod + ')', fontdict=fontdict)
        plt.xlabel('Frequency/Hz', fontdict=fontdict)
        plt.ylabel('Amplitude', fontdict=fontdict)
        plt.subplot(425)
        plt.plot(t * 1e6, th.abs(S1))
        plt.grid()
        plt.title('Matched filtered with time domain filter', fontdict=fontdict)
        plt.xlabel(r'Time/$\mu s$', fontdict=fontdict)
        plt.ylabel('Amplitude', fontdict=fontdict)
        plt.subplot(426)
        plt.plot(t * 1e6, th.abs(S2))
        plt.grid()
        plt.title('Matched filtered with frequency domain filter', fontdict=fontdict)
        plt.xlabel(r'Time/$\mu s$', fontdict=fontdict)
        plt.ylabel('Amplitude', fontdict=fontdict)
        plt.subplot(427)
        plt.plot(t * 1e6, th.angle(S1))
        plt.grid()
        plt.title('Matched filtered with time domain filter', fontdict=fontdict)
        plt.xlabel(r'Time/${\mu s}$', fontdict=fontdict)
        plt.ylabel('Phase/rad', fontdict=fontdict)
        plt.subplot(428)
        plt.plot(t * 1e6, th.angle(S2))
        plt.grid()
        plt.title('Matched filtered with frequency domain filter', fontdict=fontdict)
        plt.xlabel(r'Time/${\mu s}$', fontdict=fontdict)
        plt.ylabel('Phase/rad', fontdict=fontdict)
        plt.subplots_adjust(left=0.06, bottom=0.06, right=0.99, top=0.96, wspace=0.14, hspace=0.60)
        plt.savefig('PulseCompressionDemoThreeTargetsKoff' + str(Koff) + '.pdf')

        plt.show()


    """

def chirp_mf_fd(K, Tp, Fs, Fc=0., Nfft=None, mod='way1', win=None, ftshift=False, scale=False, device='cpu'):
    """Summary

    Parameters
    ----------
    K : int
        The chirp rate.
    Tp : float
        The pulse width.
    Fs : float
        The sampling rate.
    Fc : float, optional
        The center frequency.
    Nfft : int or None, optional
        The number of points for doing FFT.
    mod : str, optional
        The mode of matched filter.
    win : tensor or None, optional
        The window function.
    ftshift : bool, optional
        Shift the zero-frequecy in center?
    scale : bool, optional
        Scale the filter?
    device : str, optional
        Specifies the device to be used for computing.

    Returns
    -------
    tensor
        The matched filter tensor.
    """





#!/usr/bin/env python
#-*- coding: utf-8 -*-
# @file      : fast_fourier_domain_optimization.py
# @author    : Zhi Liu
# @email     : zhiliu.mind@gmail.com
# @homepage  : http://iridescent.ink
# @date      : Sun Nov 27 2022
# @version   : 2.0
# @license   : The Apache License 2.0
# @note      : 
# 
# The Apache 2.0 License
# Copyright (C) 2013- Zhi Liu
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#

from __future__ import print_function
import torch as th
import torchbox as tb
import torchsar as ts


def _compensation(X, phi, Na, Nr, ftshift=False, isfft=True):
    d = X.dim()
    sizea, sizer = [1] * d, [1] * d
    returns = []
    if Na is not None:
        pa = phi[0]
        sizea[0], sizea[-3], sizea[-1] = pa.size(0), pa.size(1), 2
        epa = th.stack((th.cos(pa), -th.sin(pa)), dim=-1)
        epa = epa.reshape(sizea)

        if isfft:
            X = tb.fft(X, cdim=-1, dim=-3, keepcdim=True, shift=ftshift)
        X = tb.ematmul(X, epa)
        returns.append(pa.data)

    if Nr is not None:
        if Na is None:
            pr = phi[0]
        else:
            pr = phi[1]
        sizer[0], sizer[-2], sizer[-1] = pr.size(0), pr.size(1), 2
        epr = th.stack((th.cos(pr), -th.sin(pr)), dim=-1)
        epr = epr.reshape(sizer)

        if isfft:
            X = tb.fft(X, cdim=-1, dim=-2, keepcdim=True, shift=ftshift)
        X = tb.ematmul(X, epr)
        returns.append(pr.data)
    return [X] + returns


class AutoFocusFFO(th.nn.Module):

    def __init__(self, Na, Nr, Nb=None, ftshift=False, trainable=True):
        super(AutoFocusFFO, self).__init__()

        if Nb is None:
            Nb = 1

        self.Na = Na
        self.Nr = Nr
        self.Nb = Nb
        self.ftshift = ftshift

        self.barephi = ts.BarePhi(Na, Nr, Nb, pa=None, pr=None, Ma=7, Mr=7, shift=ftshift, trainable=trainable)

    def forward(self, X, isfft=True):

        phi = self.barephi()
        returns = _compensation(X, phi, self.Na, self.Nr, ftshift=self.ftshift, isfft=True)

        return returns

    def imaging(self, Xc):
        if self.Na is not None:
            Xc = tb.ifft(Xc, axis=-3, shift=self.ftshift)
        if self.Nr is not None:
            Xc = tb.ifft(Xc, axis=-2, shift=self.ftshift)
        return Xc

    def get_param(self):
        param = []
        if self.barephi.pa is not None:
            param.append(self.barephi.pa)
        if self.barephi.pr is not None:
            param.append(self.barephi.pr)
        return param

    def param_init(self, pa=None, pr=None):
        self.barephi.param_init(pa, pr)




def _focusing(X, phi, Na, Nr, ftshift=False, isfft=True):
    ...

class AutoFocusBarePhi(th.nn.Module):
    ...

    def __init__(self, Na, Nr, Nb=None, pa=None, pr=None, Ma=None, Mr=None, xa=None, xr=None, ftshift=False, trainable=True):
        ...

    def forward(self, X, isfft=True):
        ...

    def get_param(self):
        ...

    def param_init(self, pa=None, pr=None):
        ...

class AutoFocusPolyPhi(th.nn.Module):
    ...

    def __init__(self, Na, Nr, Nb=None, ca=None, cr=None, Ma=2, Mr=2, xa=None, xr=None, ftshift=False, trainable=True):
        ...

    def forward(self, X, isfft=True):
        ...

    def get_param(self):
        ...

    def param_init(self, ca=None, cr=None):
        ...

class AutoFocusDctPhi(th.nn.Module):
    ...

    def __init__(self, Na, Nr, Nb=None, ca=None, cr=None, Pa=4, Pr=4, ftshift=False, trainable=True):
        ...

    def forward(self, X, isfft=True):
        ...

    def get_param(self):
        ...

    def param_init(self, ca=None, cr=None):
        ...

class AutoFocusSinPhi(th.nn.Module):
    ...

    def __init__(self, Na, Nr, Nb=None, aa=None, fa=None, ar=None, fr=None, ftshift=False, trainable=True):
        ...

    def forward(self, X, isfft=True):
        ...

    def get_param(self):
        ...

    def param_init(self, aa=None, fa=None, ar=None, fr=None):




class RangeMatchedFilter(th.nn.Module):
    ...

    def __init__(self, Na, Tp, Fsr, Kr, Fc, trainable=True, dtype=th.float64):
        ...

    def forward(self):
        ...

class AzimuthMatchedFilter(th.nn.Module):
    ...

    def __init__(self, Nr, Tp, Fsa, Ka, Fc, trainable=True, dtype=th.float32):
        ...

    def forward(self):
        ...

class AzimuthMatchedFilterLinearFit(th.nn.Module):
    ...

    def __init__(self, Nr, Tp, Fsa, Ka, Fc, trainable=True, dtype=th.float32):
        ...

    def forward(self):




def chirp_tran(t, Tp, K, Fc, a=1.):
    ...

def chirp_recv(t, Tp, K, Fc, a=1., g=1., r=1e3):
    ...

class Chirp(th.nn.Module):
    ...

    def __init__(self, Tp, K, Fc=0., a=1.):
        ...

    def tran(self, t):
        ...

    def recv(self, t, g, r):




def dsm2sar_fd(g, pdict, mode='StationaryTarget2D', ftshift=True, device='cpu'):
    """Frequency-domain simulation

    SAR raw data simulation by frequency-domain method.

    Parameters
    ----------
    g : tensor
        The digital scene matrix.
    pdict : dict
        The SAR platform parameters.
    mode : str, optional
        Simulation mode.
    ftshift : bool, optional
        Shift zero-frequency to center?
    device : str, optional
        Specifies which device to be used.

    Returns
    -------
    tensor
        Simulated SAR raw data.
    """





#!/usr/bin/env python
#-*- coding: utf-8 -*-
# @file      : slant_ground_range.py
# @author    : Zhi Liu
# @email     : zhiliu.mind@gmail.com
# @homepage  : http://iridescent.ink
# @date      : Sun Nov 27 2022
# @version   : 0.0
# @license   : The Apache License 2.0
# @note      : 
# 
# The Apache 2.0 License
# Copyright (C) 2013- Zhi Liu
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#

import math
import torch as th
from torchbox.utils.const import *


def slantr2groundr(R, H, Ar, Xc):
    """slant range to ground range

    Convert slant range :math:`R` to ground range :math:`X`.

    Parameters
    ----------
    R : 1d-tensor
        slant range array
    H : float
        sarplat height
    Ar : float
        squint angle (unit, rad) in line geometry

    Returns
    -------
    X : 1d-tensor
        ground range

    """
    return th.sqrt(th.abs((R * math.cos(Ar))**2 - H * H) + EPS) - Xc
    # return th.sqrt(th.abs((R * math.cos(Ar))**2 - H * H) + EPS)


def slantt2groundr(tr, H, Ar):
    """slant time to ground range

    Convert slant range time :math:`t_r` to ground range :math:`X`.

    Parameters
    ----------
    tr : 1d-tensor
        range time array
    H : float
        sarplat height
    Ar : float
        squint angle (unit, rad) in line geometry

    Returns
    -------
    X : 1d-tensor
        ground range

    """

    return th.sqrt(th.abs(((tr * C / 2.) * math.cos(Ar))**2 - H * H) + EPS)


def groundr2slantr(X, H, Ar, Xc):
    """ground range to slant range

    Convert ground range :math:`R` to slant range :math:`X`.

    Parameters
    ----------
    X : 1d-tensor
        ground range array
    H : float
        sarplat height
    Ar : float
        squint angle (unit, rad) in line geometry

    Returns
    -------
    R : 1d-tensor
        slant range

    """

    return th.sqrt((X + Xc)**2 + H * H + EPS) / (math.cos(Ar) + EPS)
    # return th.sqrt((X)**2 + H * H + EPS) / (math.cos(Ar) + EPS)


def groundr2slantt(X, H, Ar):
    """ground range to slant time

    Convert ground range :math:`X` to slant time :math:`t_r`.

    Parameters
    ----------
    X : 1d-tensor
        ground range
    H : float
        sarplat height
    Ar : float
        squint angle (unit, rad) in line geometry

    Returns
    -------
    tr : 1d-tensor
        range time array

    """

    return 2. * th.sqrt(X**2 + H * H + EPS) / (math.cos(Ar) + EPS) / C


def min_slant_range(Rnear, Fsr, Noff):
    """minimum slant range from radar to target

    Compute the minimum slant range from radar to target.

    Parameters
    ----------
    Rnear : float
        The nearest range (start sampling) from radar to the target.
    Fsr : float
        Sampling rate in range direction
    Noff : 1d-tensor
        Offset from the start distance (start sampling) cell.

    Returns
    -------
    r : 1d-tensor
        Minimum slant range of each range cell.
    """
    r = Rnear + Noff * (C / (2. * Fsr))
    return r


def min_slant_range_with_migration(Rnear, Fsr, Noff, Wl, Vr, fdc):
    r = Rnear + Noff * (C / (2. * Fsr))
    return r / th.sqrt(1. - (Wl * fdc / (Vr + Vr))**2)



#!/usr/bin/env python
#-*- coding: utf-8 -*-
# @file      : geometry.py
# @author    : Zhi Liu
# @email     : zhiliu.mind@gmail.com
# @homepage  : http://iridescent.ink
# @date      : Sun Nov 27 2022
# @version   : 0.0
# @license   : The Apache License 2.0
# @note      : 
# 
# The Apache 2.0 License
# Copyright (C) 2013- Zhi Liu
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#

import numpy as np


def rectangle(SceneArea, x0, y0, h, w, a=None, dx=None, dy=None, verbose=False):
    r"""rectangle area

    generates rectangle area

    Parameters
    ----------
    SceneArea : list or tuple
        Scene Area, unit m, [xmin,xmax,ymin,ymax]
    x0 : float
        x center of rectangle, unit m (x: horizontal, y: vertical)
    y0 : float
        y center of rectangle, unit m (x: horizontal, y: vertical)
    h : float
        height of rectangle, unit m (the default is None, which generates randomly)
    w : float
        width of rectangle, unit m (the default is None, which generates randomly)
    a : float, optional
        amplitude (the default is None, which generates randomly)
    dx : float, optional
        resolution in range (default: {1 / (xmax-xmin)})
    dy : float, optional
        resolution in azimuth (default: {1 / (ymax-ymin)})
    verbose : bool, optional
        show more log info (the default is False, which means does not show)

    Returns
    -------
    rects : tensor
        rectangle area
    """

    if verbose:
        print(SceneArea)
    (Xmin, Xmax, Ymin, Ymax) = SceneArea
    if dx is None:
        dx = 1. / (Xmax - Xmin)
    if dy is None:
        dy = 1. / (Ymax - Ymin)

    if h is None:
        h = np.random.rand() * 100.
    if w is None:
        w = np.random.rand() * 100.

    if a is not None:
        g = a.copy()
    rects = []
    for x in np.linspace(max(x0 - w / 2., Xmin), min(x0 + w / 2., Xmax), int(w)):
        for y in np.linspace(max(y0 - h / 2., Ymin), min(y0 + h / 2., Ymax), int(h)):
            if a is None:
                g = np.random.rand()
            rects.append([x, y, 0, 0, 0, 0, g])

    rects = np.array(rects)

    if verbose:
        print(rects, rects.shape)
    return rects


def disc(SceneArea, x0, y0, r, a=None, dx=None, dy=None, verbose=False):
    r"""disc area

    generates disc area


    Parameters
    ----------
    SceneArea : list or tuple
        Scene Area, [xmin,xmax,ymin,ymax]
    x0 : int
        x center of disc (x: horizontal, y: vertical)
    y0 : int
        y center of disc (x: horizontal, y: vertical)
    r : int
        radius of disc
    a : int, optional
        amplitude of disc (default: {None})
    dx : float, optional
        resolution in range (default: {1 / (xmax-xmin)})
    dy : float, optional
        resolution in azimuth (default: {1 / (ymax-ymin)})
    verbose : bool, optional
        show more log info (the default is False, which means does not show)

    Returns
    -------
    ccas : tensor
        disk area
    """

    if verbose:
        print(SceneArea)
    (Xmin, Xmax, Ymin, Ymax) = SceneArea
    if dx is None:
        dx = 1. / (Xmax - Xmin)
    if dy is None:
        dy = 1. / (Ymax - Ymin)
    # print("disc: x0, y0, r, a, dx, dy --> ", x0, y0, r, a, dx, dy)
    r2 = r * r
    ccas = []
    if a is not None:
        g = a.copy()
    for x in np.linspace(max(x0 - r, Xmin), min(x0 + r, Xmax), int(2 * r + 2)):
        for y in np.linspace(max(y0 - r, Ymin), min(y0 + r, Ymax), int(2 * r + 2)):
            d = (x - x0) ** 2 + (y - y0) ** 2
            if d <= r2:
                if a is None:
                    g = np.random.rand()
                ccas.append([x, y, 0, 0, 0, 0, g])

    ccas = np.array(ccas)
    if verbose:
        print(ccas, ccas.shape, "===++===")
    return ccas




#!/usr/bin/env python
#-*- coding: utf-8 -*-
# @file      : simulation_time_domain.py
# @author    : Zhi Liu
# @email     : zhiliu.mind@gmail.com
# @homepage  : http://iridescent.ink
# @date      : Sun Nov 27 2022
# @version   : 0.0
# @license   : The Apache License 2.0
# @note      : 
# 
# The Apache 2.0 License
# Copyright (C) 2013- Zhi Liu
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#

import math
import torch as th
import torchbox as tb
import torchsar as ts


def tgs2sar_td(targets, pdict, mode='Target', device='cpu'):
    """Time-domain simulation

    SAR raw data simulation by time-domain method.

    Parameters
    ----------
    targets : tensor
        A 2d-tensor contains the information of targets,
        each row is a target [x, y, z, vx, vy, vz, a].
    pdict : dict
        The SAR platform parameters.
    mode : str, optional
        Simulation mode.
    device : str, optional
        Specifies which device to be used.

    Returns
    -------
    tensor
        Simulated SAR raw data tensor.
    """
    # target --> [x, y, z, vx, vy, vz, a]
    platform = th.tensor(pdict['PlatCenter'], device=device)
    SC = th.tensor(pdict['SceneCenter'], device=device)
    if type(targets) is not th.Tensor:
        targets = th.tensor(targets)
    Kr, Tp = pdict['Kr'], pdict['Tp']
    Na, Nr = pdict['EchoSize']
    if pdict['GeometryMode'] == 'SceneGeometry':
        Rc = pdict['Rsc']
        R0 = pdict['Rs0']
    if pdict['GeometryMode'] == 'BeamGeometry':
        Rc = pdict['Rbc']
        R0 = pdict['Rb0']
    Wl, Fc, La = pdict['Wl'], pdict['Fc'], pdict['La']

    V = pdict['V']
    Vg = pdict['Vg']
    Vr = pdict['Vr']
    Ar = pdict['Ar']
    targets = targets.to(device)
    eta = pdict['ta'].to(device)
    tau = pdict['tr'].to(device)

    if th.is_complex(targets):
        rcss = targets[:, [-1]]
        targets = targets[:, 0:-1].real
    else:
        rcss = targets[:, [-1]]
        targets = targets[:, 0:-1]

    raise(ImportError('Not opened yet!'))

